from __future__ import annotations

import ast
from typing import TYPE_CHECKING

from rattr.ast.types import AstLiterals, Identifier
from rattr.config import Config
from rattr.models.symbol._symbols import (
    PYTHON_BUILTINS,
    PYTHON_NON_PRIMITIVE_RETURNING_BUILTINS,
    Import,
)

if TYPE_CHECKING:
    from rattr.models.symbol._symbol import Symbol


def is_call_to_literal(name: Identifier) -> bool:
    """Return `True` if this appears to be a call to a literal."""
    return name.startswith("@")


def is_call_to_subscript_item(name: Identifier) -> bool:
    """Return `True` if this appears to be a call to a literal."""
    return "[]" in name


def is_call_to_method(
    target: Symbol | None,
    name: Identifier,
    lhs_target: Symbol | None,
    lhs_name: Identifier,
) -> bool:
    """Return `True` if this appears to be a call to a method."""
    return name != lhs_name and target is None and not isinstance(lhs_target, Import)


def is_call_to_member_of_module_import(
    name: Identifier,
    target: Symbol | None,
) -> bool:
    """Return `True` if this appears to be a call to a module member.

    Only applicable to members of modules imported like so:
    >>> import math

    Not those imported like so:
    >>> from math import pi

    Examples:
    >>> # Context from `import math`
    >>> # thus `math.pi` is not explicitly defined
    >>> # thus the rhs is `None`
    >>> is_call_to_member_of_module_import("math.pi", None)
    True

    >>> # Context from `from math import pi`
    >>> # thus `math.pi` is defined in the context
    >>> # thus the rhs is the import symbol for `pi`
    >>> is_call_to_member_of_module_import("pi", Import(...))
    False
    """
    return "." in name and target is None


def is_direct_call_to_method_on_constant(name: str) -> bool:
    """Return `True` if the name is a call to method on a constant."""
    config = Config()

    _prefix = config.LITERAL_VALUE_PREFIX
    _constant = f"{_prefix}{ast.Constant.__name__}."

    return name.startswith(_constant)


def is_direct_call_to_method_on_literal(name: str) -> bool:
    """Return `True` if the name is a call to method on a constant."""
    config = Config()

    _prefix = config.LITERAL_VALUE_PREFIX
    _literals = tuple(f"{_prefix}{literal.__name__}." for literal in AstLiterals)

    return name.startswith(_literals)


def is_call_to_method_on_primitive_from_call(name: str) -> bool:
    """Return `True` if the name is a call to a method on a call result's primitive.

    This will have lots of false-negatives, but it will detect calls to methods on the
    results of casts or other applicable builtins.
    """
    builtins = {b for b in PYTHON_BUILTINS if b[0].isalpha()}
    builtins -= PYTHON_NON_PRIMITIVE_RETURNING_BUILTINS

    return any(name.startswith((f"{b}.", f"{b}().")) for b in builtins)


def is_call_to_method_on_py_type(name: str) -> bool:
    """Return `True` if the given name is a method on a primitive type."""
    if is_direct_call_to_method_on_constant(name):
        return True

    if is_direct_call_to_method_on_literal(name):
        return True

    if is_call_to_method_on_primitive_from_call(name):
        return True

    return False


def is_call_to_call_result(call: ast.expr) -> bool:
    """Return `True` if the given call is a call to another call's result."""
    return isinstance(call, ast.Call) and isinstance(call.func, ast.Call)



from enum import Enum

from rattr.analyser.annotations import rattr_results

from .first import some_function
from .module.foo import my_lib_function


def recursive(x):
    if not x.xs:
        return 0

    if len(x.xs) == 1:
        return x.xs[0].single_value

    return sum(_x.multivalue for _x in x.xs)


def _one_fn(arg_a, arg_b):
    return arg_a.attr + some_function(arg_b.some_thing, 1, 2)


def _two_fn(first, second):
    return _one_fn(second, first)


def _three_fn(arg_a, arg_b):
    return sum(
        [
            (getattr(arg_a, "first", 0) * getattr(arg_b, "first", 0))
            + (getattr(arg_a, "second", 0) * getattr(arg_b, "second", 0))
            + (getattr(arg_a, "third", 0) * getattr(arg_b, "third", 0))
        ]
    )


def _four_fn(arg_a, arg_b=None):
    try:
        my_lib_function(arg_a)
    except Exception:
        return 0
    return arg_a.accessed_in_four


class Things(Enum):
    one = 1
    two = 2
    three = 3


_functions_by_thing = {
    Things.one: _one_fn,
    Things.two: _two_fn,
    Things.three: _three_fn,
    # Things.four: _four_fn,  # this is skipped but will be in results, deliberately
}


@rattr_results(
    gets={
        "data.discriminator",
        "data.wrapped",
    },
    calls=[
        ("_one_fn()", (["data.wrapped"], {})),
        ("_two_fn()", (["data.wrapped"], {})),
        ("_three_fn()", (["data.wrapped"], {})),
        ("_four_fn()", (["data.wrapped"], {})),
    ],
)
def dispatcher(data):
    return _functions_by_thing[data.discriminator](data.wrapped)



from __future__ import annotations

import abc
import ast
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple, Union

import attrs
from attrs import field
from frozendict import frozendict

from rattr.config.util import get_current_file
from rattr.models.symbol._util import PYTHON_BUILTINS_LOCATION, arg_name, kwarg_name

if TYPE_CHECKING:
    from typing import Literal

    from rattr.ast.types import Identifier


@attrs.frozen
class Symbol(abc.ABC):
    name: str = field()

    token: Union[ast.AST, None] = field(
        default=None,
        kw_only=True,
        hash=False,
        eq=False,
    )
    location: Union[Location, None] = field(
        default=None,
        kw_only=True,
        hash=False,
        eq=False,
    )

    interface: Union[CallInterface, None] = field(default=None, kw_only=True)

    def __attrs_pre_init__(self) -> None:
        if type(self) == Symbol:
            raise NotImplementedError("symbol should be sub-classed")

    @property
    def id(self) -> str:
        # For most symbols the name is already the identifier, but for starred-imports
        # we must prepend the qualified name to avoid namespace collisions.
        # See: Import.id(...)
        return self.name

    @property
    def is_callable(self) -> bool:
        return self.interface is not None

    @property
    def has_location(self) -> bool:
        return (
            self.location is not None
            and str(self.location.defined_in) != PYTHON_BUILTINS_LOCATION
        )

    @property
    def is_import(self) -> Literal[False]:
        return False

    def __lt__(self, other: Symbol | object) -> bool:
        if not isinstance(other, Symbol):
            raise TypeError
        return self.name < other.name


class ConsumableCallInterface(NamedTuple):
    posonlyargs: list[Identifier]
    args: list[Identifier]
    vararg: Identifier | None
    kwonlyargs: list[Identifier]
    kwarg: Identifier | None


@attrs.frozen
class CallInterface:
    # Naming inherited from `ast.arguments`
    posonlyargs: tuple[str, ...] = field(default=(), converter=tuple)
    args: tuple[str, ...] = field(default=(), converter=tuple)
    vararg: Union[str, None] = field(default=None)
    kwonlyargs: tuple[str, ...] = field(default=(), converter=tuple)
    kwarg: Union[str, None] = field(default=None)

    @property
    def all(self) -> tuple[str, ...]:
        arguments: list[str] = []

        arguments += self.posonlyargs
        arguments += self.args

        if self.vararg is not None:
            arguments.append(self.vararg)

        arguments += self.kwonlyargs

        if self.kwarg is not None:
            arguments.append(self.kwarg)

        return tuple(arguments)

    @classmethod
    def from_fn_def(
        cls: type[CallInterface],
        fn: ast.Lambda | ast.FunctionDef | ast.AsyncFunctionDef,
    ) -> CallInterface:
        """Return a new `CallInterface` parsed from the given function def."""
        return cls.from_arguments(fn.args)

    @classmethod
    def from_arguments(
        cls: type[CallInterface],
        arguments: ast.arguments,
    ) -> CallInterface:
        """Return a new `CallInterface` parsed from the given arguments."""
        if isinstance(_vararg := arguments.vararg, ast.arg):
            vararg = _vararg.arg
        else:
            vararg = None

        if isinstance(_kwarg := arguments.kwarg, ast.arg):
            kwarg = _kwarg.arg
        else:
            kwarg = None

        return CallInterface(
            posonlyargs=[_a.arg for _a in arguments.posonlyargs],
            args=[_arg.arg for _arg in arguments.args],
            vararg=vararg,
            kwonlyargs=[_arg.arg for _arg in arguments.kwonlyargs],
            kwarg=kwarg,
        )

    def as_consumable_call_interface(self) -> ConsumableCallInterface:
        """Return a copy of this interface as a consumable named tuple.

        This is s.t. the caller can freely `pop` or otherwise destructively consume the
        interface at the point of call.
        """
        return ConsumableCallInterface(
            posonlyargs=[arg for arg in self.posonlyargs],
            args=[arg for arg in self.args],
            vararg=self.vararg,
            kwonlyargs=[kwarg for kwarg in self.kwonlyargs],
            kwarg=self.kwarg,
        )


@attrs.frozen
class AnyCallInterface(CallInterface):
    """Denote that the interface is unknown/unknowable; thus all calls are accepted."""

    def __init__(self) -> None:
        super().__init__()

    @classmethod
    def from_fn_def(
        cls: type[CallInterface],
        fn: ast.Lambda | ast.FunctionDef | ast.AsyncFunctionDef,
    ) -> CallInterface:
        if cls != CallInterface:
            raise NotImplementedError(f"not applicable to {cls.__name__}")

    @classmethod
    def from_arguments(
        cls: type[CallInterface],
        arguments: ast.arguments,
    ) -> CallInterface:
        if cls != CallInterface:
            raise NotImplementedError(f"not applicable to {cls.__name__}")


@attrs.frozen
class CallArguments:
    args: tuple[str, ...] = field(default=(), converter=tuple)
    kwargs: frozendict[str, str] = field(factory=frozendict, converter=frozendict)

    @classmethod
    def from_call(
        cls: type[CallArguments],
        call: ast.Call,
        *,
        self: Identifier | None = None,
    ) -> CallArguments:
        """Return a new `CallArguments` parsed from the given function call."""
        args: list[str] = [
            arg_name(
                arg,
                culprit=call,
            )
            for arg in call.args
        ]
        kwargs: dict[str, str] = {
            kw.arg: kwarg_name(
                kw,
                culprit=call,
            )
            for kw in call.keywords
        }

        if self is not None:
            args = [self, *args]

        return CallArguments(args=args, kwargs=kwargs)


@attrs.frozen
class Location:
    lineno: int = field()
    col_offset: int = field()

    end_lineno: Union[int, None] = field(default=None)
    end_col_offset: Union[int, None] = field(default=None)

    _file: Path = field(alias="file", factory=get_current_file, kw_only=True)

    @classmethod
    def from_ast_token(cls, token: ast.AST, *, file: Path | None = None) -> Location:
        kwargs = {
            "lineno": token.lineno,
            "end_lineno": token.end_lineno,
            "col_offset": token.col_offset,
            "end_col_offset": token.end_col_offset,
        }

        if file is not None:
            kwargs["file"] = file

        return Location(**kwargs)

    @property
    def file(self) -> Path:
        return self._file

    @property
    def defined_in(self) -> Path:
        return self._file



from __future__ import annotations

import ast
from typing import TYPE_CHECKING

import pytest

from rattr.models.context import Context
from rattr.models.symbol import CallInterface, Func, Name

if TYPE_CHECKING:
    from tests.shared import ArgumentsFn, MakeSymbolTableFn, ParseFn


def test_context_add():
    root = Context(parent=None)
    child = Context(parent=root)

    root.add(symbol_in_root := Name(name="name_in_root"))
    child.add(symbol_in_child := Name(name="name_in_child"))

    assert symbol_in_root.name in root
    assert symbol_in_root.name in child

    assert symbol_in_child.name not in root
    assert symbol_in_child.name in child


def test_context_add_does_not_duplicate():
    root = Context(parent=None)
    child = Context(parent=root)

    root.add(symbol := Name(name="name"))
    child.add(symbol)

    assert symbol in root.symbol_table.symbols

    assert symbol not in child.symbol_table.symbols
    assert symbol.name in child


def test_context_add_multiple():
    root = Context(parent=None)
    symbols = (Name("first"), Name("second"), Name("third"))

    root.add(symbols)

    for symbol in symbols:
        assert symbol.name in root


def test_context_add_as_argument(make_symbol_table: MakeSymbolTableFn):
    # When added as an argument a symbol should be in both the root and child symbol
    # tables, when normally it would not be re-added.
    # See test_context_add_does_not_duplicate
    root = Context(parent=None)
    child = Context(parent=root)

    root.add(symbol := Name(name="name"))
    child.add(symbol, is_argument=True)

    expected_symbol_table = make_symbol_table({symbol.name: symbol})
    assert root.symbol_table == expected_symbol_table
    assert child.symbol_table == expected_symbol_table


def test_context_remove():
    root = Context(parent=None)

    root.add((a := Name("a"), b := Name("b")))
    assert root.symbol_table._symbols == {a.name: a, b.name: b}
    assert a.name in root

    root.remove(a.name)
    assert root.symbol_table._symbols == {b.name: b}
    assert a.name not in root


def test_context_remove_from_parent():
    root = Context(parent=None)
    root.add((a := Name("a"), b := Name("b")))

    child = Context(parent=root)

    assert root.symbol_table._symbols == {a.name: a, b.name: b}
    assert child.symbol_table._symbols == {}

    child.remove(a.name)
    assert root.symbol_table._symbols == {a.name: a, b.name: b}
    assert child.symbol_table._symbols == {}

    root.remove(a.name)
    assert root.symbol_table._symbols == {b.name: b}
    assert child.symbol_table._symbols == {}


def test_context_remove_non_existent():
    root = Context(parent=None)
    root.add((a := Name("a"), b := Name("b")))

    child = Context(parent=root)

    root.remove("fake")
    root.remove("false")

    assert root.symbol_table._symbols == {a.name: a, b.name: b}
    assert child.symbol_table._symbols == {}


def test_context_remove_all():
    root = Context(parent=None)
    root.add((a := Name("a"), b := Name("b")))

    root.remove((a.name, b.name, "fake"))
    assert root.symbol_table._symbols == {}


def test_context_get():
    root = Context(parent=None)
    root.add((a := Name("a"), b := Name("b")))

    assert root.get(a.name) == a
    assert root.get(b.name) == b
    assert root.get("fake") is None


def test_context_get_call_target_not_defined(capfd: pytest.CaptureFixture[str]):
    root = Context(parent=None)
    assert root.get_call_target("anything", culprit=None) is None

    _, stderr = capfd.readouterr()
    assert "unable to resolve call to 'anything()', target is undefined" in stderr


def test_context_get_call_target_not_callable(capfd: pytest.CaptureFixture[str]):
    root = Context(parent=None)
    root.add(a := Name("a"))
    child = Context(parent=root)

    # The root defines 'a' so it is likely a procedural parameter in practise
    assert root.get_call_target(a.name, culprit=None) == a
    _, stderr = capfd.readouterr()
    assert (
        "unable to resolve call to 'a()', target is likely a procedural parameter"
        in stderr
    )

    # The child merely inherits 'a' so it is just generically not callable
    assert child.get_call_target(a.name, culprit=None) == a
    _, stderr = capfd.readouterr()
    assert "unable to resolve call to 'a()', target is not callable" in stderr


def test_context_get_call_target_method(
    capfd: pytest.CaptureFixture[str],
    arguments: ArgumentsFn,
):
    root = Context(parent=None)
    root.add(Name("a"))

    with arguments(_warning_level="all"):
        assert root.get_call_target("a.method()", culprit=None) is None

    _, stderr = capfd.readouterr()
    assert "unable to resolve call to 'a.method()', target is a method" in stderr


def test_context_get_call_func_target(capfd: pytest.CaptureFixture[str]):
    root = Context(parent=None)
    root.add(func := Func(name="a", interface=CallInterface()))

    assert root.get_call_target("a()", culprit=None) == func


def test_context_get_call_target_brackets_dont_matter():
    root = Context(parent=None)
    root.add(a := Func(name="a", interface=CallInterface()))

    assert root.get_call_target(a.name, culprit=None) == a
    assert root.get_call_target(f"{a.name}()", culprit=None) == a


def test_context_declares():
    root = Context(parent=None)
    root.add(a := Name("a"))

    child = Context(root)
    child.add(b := Name("b"))

    assert root.declares(a.name)
    assert not child.declares(a.name)

    assert not root.declares(b.name)
    assert child.declares(b.name)


def test_context_in():
    root = Context(parent=None)
    root.add(a := Name("a"))

    child = Context(root)
    child.add(b := Name("b"))

    assert a.name in root
    assert b.name not in root

    assert a.name in child
    assert b.name in child


def test_context_add_identifiers_to_context_single():
    expr: ast.Name = ast.parse("a = 1").body[0].targets[0]

    root = Context(parent=None)
    root.add_identifiers_to_context(expr)

    assert root.symbol_table._symbols == {"a": Name("a", token=expr)}


def test_context_add_identifiers_to_context_multiple():
    expr: ast.Name = ast.parse("a, b, c = 1, 2, 3").body[0].targets[0]

    root = Context(parent=None)
    root.add_identifiers_to_context(expr)
    assert root.symbol_table._symbols == {
        "a": Name("a", token=expr),
        "b": Name("b", token=expr),
        "c": Name("c", token=expr),
    }


def test_context_remove_identifiers_from_context_single():
    expr: ast.Name = ast.parse("del a").body[0].targets[0]

    root = Context(parent=None)

    root.add(Name("a"))
    assert root.symbol_table._symbols == {"a": Name("a")}

    root.remove_identifiers_from_context(expr)
    assert root.symbol_table._symbols == {}


def test_context_remove_identifiers_from_context_multiple():
    targets: list[ast.Name] = ast.parse("del a, b").body[0].targets
    (expr_a, expr_b) = targets

    root = Context(parent=None)
    root.add(Name("a"))
    root.add(Name("b"))
    assert root.symbol_table._symbols == {"a": Name("a"), "b": Name("b")}

    root.remove_identifiers_from_context(expr_a)
    root.remove_identifiers_from_context(expr_b)
    assert root.symbol_table._symbols == {}


@pytest.mark.parametrize(
    "source_code, expected_names",
    [
        (
            """
            def fn():
                pass
            """,
            [],
        ),
        (
            """
            def fn(a, b, c=1, d=2, *x, **y):
                pass
            """,
            ["a", "b", "c", "d", "x", "y"],
        ),
    ],
)
def test_context_add_arguments_to_context(
    source_code: str,
    expected_names: list[str],
    parse: ParseFn,
    make_symbol_table: MakeSymbolTableFn,
):
    function: ast.FunctionDef = parse(source_code).body[0]

    root = Context(parent=None)
    root.add_arguments_to_context(function.args, token=None)

    assert root.symbol_table == make_symbol_table(
        [Name(name, token=function) for name in expected_names]
    )


def test_context_redefinition_in_scope():
    root = Context(parent=None)

    root.add(Name("var_one"))
    root.add(Name("var_one"))

    assert root.symbol_table._symbols == {"var_one": Name("var_one")}


def test_context_redefinition_in_child_scope():
    root = Context(parent=None)
    root.add(Name("var_one"))

    child = Context(parent=root)
    child.add(Name("var_one"))

    assert root.symbol_table._symbols == {"var_one": Name("var_one")}
    assert child.symbol_table._symbols == {}



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

from rattr.models.symbol import Call, Name

if TYPE_CHECKING:
    from collections.abc import Iterable


class FunctionIr(TypedDict):
    gets: set[Name]
    sets: set[Name]
    dels: set[Name]
    calls: set[Call]

    @classmethod
    def the_empty_ir(cls) -> FunctionIr:
        return {"gets": set(), "sets": set(), "dels": set(), "calls": set()}

    @classmethod
    def new(
        cls,
        *,
        gets: Iterable[Name] = (),
        sets: Iterable[Name] = (),
        dels: Iterable[Name] = (),
        calls: Iterable[Call] = (),
    ) -> FunctionIr:
        return {
            "gets": set(gets),
            "sets": set(sets),
            "dels": set(dels),
            "calls": set(calls),
        }



from __future__ import annotations

import ast
from functools import lru_cache
from typing import TYPE_CHECKING

from rattr import error
from rattr.ast.types import (
    AstComprehensions,
    AstLiterals,
)
from rattr.config import Config
from rattr.models.symbol._symbols import PYTHON_ATTR_ACCESS_BUILTINS

if TYPE_CHECKING:
    from rattr.ast.types import Identifier


def is_call_to_fn(node: ast.Call, target: Identifier) -> bool:
    """Return `True` if the given node is a direct call to `target`."""
    if not isinstance(node, ast.Call):
        raise TypeError(f"line {node.lineno}: {ast.dump(node)}")

    if not isinstance(node.func, ast.Name):
        return False

    return node.func.id == target


def is_string_literal(node: ast.expr) -> bool:
    return isinstance(node, ast.Constant) and isinstance(node.value, str)


def get_python_attr_access_fn_obj_attr_pair(
    fn: str,
    node: ast.Call,
    *,
    warn: bool = False,
) -> tuple[Identifier, Identifier]:
    """Return the object-attr pair from the call to getattr, etc.

    ### Examples
    >>> # Simple getattr call
    >>> call = ast.parse("getattr(a, 'b')").body[0].value
    >>> get_python_attr_access_fn_obj_name_pair("getattr", call)
    ("a", "b")

    >>> # Nested getattr call
    >>> call = ast.parse("getattr(getattr(a, 'b'), 'c')").body[0].value
    >>> get_python_attr_access_fn_obj_name_pair("getattr", call)
    ("a.b", "c")

    >>> # Complex nested getattr call
    >>> call = ast.parse("getattr(getattr(a.b[0], 'c'), 'd')").body[0].value
    >>> get_python_attr_access_fn_obj_name_pair("getattr", call)
    ("a.b[].c", "d")

    >>> # Non-literal in RHS
    >>> # Additionally, this will give a rattr.error
    >>> call = ast.parse("getattr(a, some_string_variable)").body[0].value
    >>> get_python_attr_access_fn_obj_name_pair("getattr", call)
    ("a", "<some_string_variable>")
    """
    this = f"{fn!r}"

    try:
        obj, name, *_ = node.args
    except ValueError:
        error.fatal(f"invalid call to {this}, too few args", node)

    if warn and not isinstance(name, ast.Constant):
        error.error(f"{this} expects 'name' to be a string literal", node)

    _, varname = names_of(name, safe=True)
    attr: str = name.value if is_string_literal(name) else f"<{varname}>"

    if isinstance(obj, ast.Call) and not is_call_to_fn(obj, fn):
        error.fatal(f"{this} may only be nested in other calls to {this}", node)

    if isinstance(obj, ast.Call):
        lhs = ".".join(get_python_attr_access_fn_obj_attr_pair(fn, obj, warn=warn))
    else:
        _, lhs = names_of(obj, safe=False)

    return lhs, attr


@lru_cache(maxsize=None)
def names_of(
    node: ast.expr,
    *,
    unravel_attr_access_calls: bool = True,
    safe: bool = False,
) -> tuple[Identifier, Identifier]:
    """Return the node's basename and fullname.

    Uses indirect recursion, see:
    * __ast_call_name
    * __ast_compound_name
    * get_python_attr_access_fn_obj_name_pair

    ### Examples
    >>> # Simple examples of strictly-nameable expressions
    >>> # These are the same for both safe=False and safe=True
    >>> names_of(ast.parse("my_variable").body[0].value)
    ("my_variable", "my_variable")
    >>> names_of(ast.parse("my_variable.attr").body[0].value)
    ("my_variable", "my_variable.attr")
    >>> names_of(ast.parse("my_variable[100].attr").body[0].value)
    ("my_variable", "my_variable[].attr")
    >>> names_of(ast.parse("*my_variable[100].attr").body[0].value)
    ("my_variable", "*my_variable[].attr")

    >>> # Not a strictly nameable expression
    >>> # With safe=False (default)
    >>> names_of(ast.parse("(a, b)[0].attr").body[0].value)
    error.RattrLiteralInNameable: ...
    >>> # With safe=True
    >>> names_of(ast.parse("(a, b)[0].attr").body[0].value, safe=True)
    ("@Tuple", "@Tuple[].attr")
    """
    if isinstance(node, ast.Name):
        return node.id, node.id

    if isinstance(node, ast.Call):
        return __ast_call_name(
            node,
            unravel_attr_access_calls=unravel_attr_access_calls,
            safe=safe,
        )

    if isinstance(
        node,
        (ast.Attribute, ast.Subscript, ast.Starred),
    ):
        return __ast_compound_name(node, safe=safe)

    # The expression is unnamable, if in safe mode give a "best-guess" as to avoid an
    # error, but when not in safe mode pick an appropriate error type.

    if safe:
        return __safe_name(node), __safe_name(node)

    raise __specific_name_error(node)(f"line {node.lineno}: {ast.dump(node)}")


@lru_cache(maxsize=1)
def __safe_name(node: ast.expr) -> Identifier:
    """Return a safe name for this unnameable expression."""
    config = Config()

    _local_value_prefix = config.LITERAL_VALUE_PREFIX
    _node_class_name = node.__class__.__name__

    return f"{_local_value_prefix}{_node_class_name}"


def __ast_call_name(
    node: ast.Call,
    *,
    unravel_attr_access_calls: bool = False,
    safe: bool = True,
) -> tuple[Identifier, Identifier]:
    basename, lhs_name = names_of(node.func, safe=safe)

    # Special case: `getattr`, etc
    if unravel_attr_access_calls and basename in PYTHON_ATTR_ACCESS_BUILTINS:
        obj, attr = get_python_attr_access_fn_obj_attr_pair(basename, node)
        return basename, f"{obj}.{attr}"

    return basename, f"{lhs_name}()"


def __ast_compound_name(
    node: ast.Attribute | ast.Starred | ast.Subscript,
    *,
    safe: bool = True,
) -> tuple[Identifier, Identifier]:
    basename, lhs_name = names_of(node.value, safe=safe)

    if isinstance(node, ast.Attribute):
        return basename, f"{lhs_name}.{node.attr}"

    if isinstance(node, ast.Subscript):
        return basename, f"{lhs_name}[]"

    if isinstance(node, ast.Starred):
        return basename, f"*{lhs_name}"

    raise NotImplementedError


def __specific_name_error(node: ast.expr) -> type[TypeError]:
    # Give a specific error if possible
    if isinstance(node, ast.UnaryOp):
        return error.RattrUnaryOpInNameable

    if isinstance(node, ast.BinOp):
        return error.RattrBinOpInNameable

    if isinstance(node, ast.Constant):
        return error.RattrConstantInNameable

    if isinstance(node, AstLiterals):
        return error.RattrLiteralInNameable

    if isinstance(node, AstComprehensions):
        return error.RattrComprehensionInNameable

    # We could not be specific :(
    return TypeError



"""Rattr function analyser."""
from __future__ import annotations

import ast
from itertools import accumulate
from typing import TYPE_CHECKING

from rattr import error
from rattr.analyser.base import NodeVisitor
from rattr.analyser.util import (
    assignment_is_one_to_one,
    class_in_rhs,
    get_assignment_targets,
    get_function_body,
    is_call_to,
    lambda_in_rhs,
    namedtuple_in_rhs,
)
from rattr.ast.types import AstFunctionDefOrLambda, AstNodeWithName
from rattr.ast.util import (
    fullname_of,
    namedtuple_init_signature_from_declaration,
    names_of,
)
from rattr.config import Config
from rattr.models.context import Context, new_context
from rattr.models.ir import FunctionIr
from rattr.models.symbol import (
    PYTHON_ATTR_ACCESS_BUILTINS,
    Call,
    CallInterface,
    Class,
    Func,
    Name,
)
from rattr.models.symbol.util import without_call_brackets
from rattr.plugins import plugins

if TYPE_CHECKING:
    from rattr.models.plugins import CustomFunctionAnalyser


def custom_analyser_for_target(
    node: ast.Call,
    context: Context,
) -> CustomFunctionAnalyser:
    target_name = without_call_brackets(
        fullname_of(
            node,
            unravel_attr_access_calls=False,
            safe=True,
        )
    )
    target_symbol = context.get_call_target(target_name, node, warn=False)

    return plugins.get_analyser(target_symbol, modulename=context.modulename)


class FunctionAnalyser(NodeVisitor):
    """Walk a function's AST to determine the accessed attributes."""

    def __init__(self, ast_function: ast.AST, context: Context) -> None:
        """Set configuration and initialise IR."""
        if not isinstance(ast_function, AstFunctionDefOrLambda):
            raise TypeError("FunctionAnalyser expects `_ast` to be a function")

        self.ast: ast.Lambda | ast.FunctionDef | ast.AsyncFunctionDef = ast_function

        self.func_ir: FunctionIr = {
            "gets": set(),
            "sets": set(),
            "dels": set(),
            "calls": set(),
        }

        self.context: Context = context
        """Managed by `new_context`."""

    def analyse(self) -> FunctionIr:
        """Entry point, return the results of analysis."""
        with new_context(self):
            self.context.add_arguments_to_context(self.ast.args, token=self.ast)

            for stmt in get_function_body(self.ast):
                self.visit(stmt)

        return self.func_ir

    def get_and_verify_name(
        self,
        node: ast.expr,
        ctx: ast.expr_context,
    ) -> tuple[str, str]:
        """Return the name, also verify validity."""
        config = Config()

        base, full = names_of(node, safe=True)

        is_undeclared = base not in self.context
        is_assignment = isinstance(ctx, ast.Store)
        is_literal = base.startswith(config.LITERAL_VALUE_PREFIX)

        if is_undeclared and not is_assignment and not is_literal:
            error.warning(f"{base!r} potentially undefined", node)

        return base, full

    def update_results(self, symbol: Name, ctx: ast.expr_context) -> None:
        """Add the given name to the result."""
        if isinstance(ctx, ast.Store):
            self.func_ir["sets"].add(symbol)

        if isinstance(ctx, ast.Load):
            self.func_ir["gets"].add(symbol)

        if isinstance(ctx, ast.Del):
            self.func_ir["dels"].add(symbol)

    # ----------------------------------------------------------------------- #
    # Result alterors: variables, call expressions, and subscripting
    # ----------------------------------------------------------------------- #

    def visit_Name(self, node: ast.Name) -> None:
        """Visit ast.Name(id: str, ctx: ast.expr_context)."""
        basename, fullname = self.get_and_verify_name(node, node.ctx)
        self.update_results(Name(fullname, basename, token=node), node.ctx)

    def visit_compound_name(
        self,
        node: ast.Attribute | ast.Starred | ast.Subscript,
    ) -> None:
        """Helper method for special nameable nodes.

        Visit ast.Starred(value, ctx).
        Visit ast.Attribute(value, attr: str, ctx: ast.expr_context).
        Visit ast.Subscript(value, slice, ctx).

        """
        basename, fullname = self.get_and_verify_name(node, node.ctx)

        # Visit the operands in a BinOp such as `(a + b).thing`, etc.
        if not isinstance(node.value, AstNodeWithName):
            self.visit(node.value)

        self.update_results(Name(fullname, basename, token=node), node.ctx)

    def visit_Starred(self, node: ast.Starred) -> None:
        self.visit_compound_name(node)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        self.visit_compound_name(node)

    def visit_Subscript(self, node: ast.Subscript) -> None:
        self.visit_compound_name(node)

    def visit_Call(self, node: ast.Call) -> None:
        """Visit ast.Call(func, args, keywords)."""
        config = Config()

        if (analyser := custom_analyser_for_target(node, self.context)) is not None:
            return self.visit_call_to_target_with_custom_analyser(node, analyser)

        _, fullname = self.get_and_verify_name(node, ast.Load())
        target = self.context.get_call_target(fullname, node, warn=True)

        # NOTE
        # Add the call to the IR and manually visit the arguments, but not `node.func`
        # as it will register an incorrect "gets"

        if isinstance(target, Class):
            error.warning(f"{target.name!r} initialised but not stored", node)
            self_name = config.LITERAL_VALUE_PREFIX + target.name
        else:
            self_name = None

        # TODO Refactor
        # On a call to `cls.member.method()` then it must get `class.member`
        parts = without_call_brackets(fullname).split(".")[:-1]
        for attr in list(accumulate(parts, lambda a, b: f"{a}.{b}"))[1:]:
            self.func_ir["gets"].add(Name(attr, parts[0], token=node))

        call = Call.from_call(fullname, call=node, target=target, self=self_name)
        self.func_ir["calls"].add(call)

        for arg in (*node.args, *node.keywords):
            self.visit(arg)

    def visit_call_to_target_with_custom_analyser(
        self,
        node: ast.Call,
        custom_analyser: CustomFunctionAnalyser,
    ) -> None:
        target_name = without_call_brackets(
            fullname_of(
                node,
                unravel_attr_access_calls=False,
                safe=True,
            )
        )
        target_call_ir = custom_analyser.on_call(target_name, node, self.context)

        self.func_ir["gets"] |= target_call_ir["gets"]
        self.func_ir["sets"] |= target_call_ir["sets"]
        self.func_ir["dels"] |= target_call_ir["dels"]
        self.func_ir["calls"] |= target_call_ir["calls"]

    # ----------------------------------------------------------------------- #
    # Context alterors: assignments and deleteion
    # ----------------------------------------------------------------------- #

    def visit_LambdaAssign(
        self,
        node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
        targets: list[ast.expr],
    ) -> None:
        """Helper method for handling non-anonymous lambdas."""
        target = targets[0]

        if not assignment_is_one_to_one(node):
            error.fatal("lambda assignment must be one-to-one", node)

        error.error("unable to unbind lambdas defined in functions", node)

        name = fullname_of(target)

        if not isinstance(node.value, ast.Lambda):
            error.fatal("unable to find lambda in rhs")  # never

        func = Func(
            name=name,
            token=node,
            interface=CallInterface.from_fn_def(node.value),
        )
        self.context.add(func)

    def visit_NamedTupleAssign(
        self,
        node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
        targets: list[ast.expr],
    ) -> None:
        """Helper method for handling non-anonymous lambdas."""
        target = targets[0]

        if not assignment_is_one_to_one(node):
            error.fatal("namedtuple assignment must be one-to-one", culprit=node)

        name = fullname_of(target)
        try:
            arguments = namedtuple_init_signature_from_declaration(node)
        except ValueError as exc:
            return error.error(str(exc.args[0]), culprit=node)

        cls = Class(name=name, token=node, interface=CallInterface(args=arguments))
        self.context.add(cls)

    def visit_ClassAssign(
        self,
        node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
        targets: list[ast.expr],
    ) -> None:
        """Helper method for assignments where RHS is a new class instance."""
        target = targets[0]

        if not assignment_is_one_to_one(node):
            error.fatal("class assignment must be one-to-one", node)

        # Create call to class initialiser
        lhs_basename, lhs_name = names_of(target)

        class_name = fullname_of(node.value)
        init_body = self.context.get_call_target(class_name, node)

        call = Call.from_call(
            class_name,
            call=node.value,
            target=init_body,
            self=lhs_name,
        )
        self.func_ir["calls"].add(call)

        # Create set to LHS
        self.func_ir["sets"].add(Name(lhs_name, lhs_basename, token=node))

        # Register assignments
        for target in targets:
            self.context.add_identifiers_to_context(target)

        # Visit call arguments
        for arg in (*node.value.args, *node.value.keywords):
            self.visit(arg)

    def visit_AnyAssign(
        self,
        node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
    ) -> None:
        """Helper method for assignments.

        Visit ast.Assign(targets, value, type_comment)\\
        Visit ast.AnnAssign(target, annotation, value, simple)\\
        Visit ast.AugAssign(target, op, value)

        """
        targets = get_assignment_targets(node)

        if lambda_in_rhs(node):
            self.visit_LambdaAssign(node, targets)
            return

        if namedtuple_in_rhs(node):
            self.visit_NamedTupleAssign(node, targets)
            return

        if class_in_rhs(node, self.context):
            self.visit_ClassAssign(node, targets)
            return

        for target in targets:
            self.context.add_identifiers_to_context(target)

        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        self.visit_AnyAssign(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_NamedExpr(self, node: ast.NamedExpr) -> None:
        self.func_ir["sets"].add(Name(*names_of(node.target), token=node))

        if lambda_in_rhs(node):
            self.visit(node.value)

        self.visit_AnyAssign(node)

    def visit_Delete(self, node: ast.Delete) -> None:
        """Visit ast.Delete(targets)."""
        for target in node.targets:
            self.context.remove_identifiers_from_context(target)

        self.generic_visit(node)

    def _visit_for_loop(self, node: ast.For | ast.AsyncFor) -> None:
        self.context.add_identifiers_to_context(node.target)
        self.generic_visit(node)

    def visit_For(self, node: ast.For) -> None:
        self._visit_for_loop(node)

    def visit_AsyncFor(self, node: ast.AsyncFor) -> None:
        self._visit_for_loop(node)

    def visit_With(self, node: ast.With) -> None:
        """Visit ast.With(items, body, type_comment)."""
        for item in node.items:
            if not item.optional_vars:
                continue
            self.context.add_identifiers_to_context(item.optional_vars)

        self.generic_visit(node)

    def visit_AsyncWith(self, node: ast.AsyncWith) -> None:
        """Visit ast.class AsyncWith(items, body)."""
        for item in node.items:
            if not item.optional_vars:
                continue
            self.context.add_identifiers_to_context(item.optional_vars)

        return self.generic_visit(node)

    # ----------------------------------------------------------------------- #
    # Context creators: function, class definitions, etc
    # ----------------------------------------------------------------------- #

    def visit_AnyFunctionDef(
        self,
        node: ast.FunctionDef | ast.AsyncFunctionDef | ast.Lambda,
    ) -> None:
        """Helper method for visiting nested function definitions.

        Visit ast.FunctionDef(name, args, body, decorator_list, returns)\\
        Visit ast.AsyncFunctionDef(name, args, body, decorator_list, returns)\\
        Visit ast.Lambda(args, body)

        NOTE:
            visit_Lambda(lambda) and thus visit_AnyFunctionDef(lambda), are
            only reached when the lambda is an anonymous function.

        """
        if not isinstance(node, ast.Lambda):
            error.error("unable to unbind nested functions", node)
        else:
            error.error("unable to unbind anonymous lambdas", node)

        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            self.context.add(Func.from_fn_def(node))

        # TODO Allow nested -- FunctionAnalyser on name "outer.inner"
        with new_context(self):
            self.context.add_arguments_to_context(node.args, token=node)

            for stmt in get_function_body(node):
                self.visit(stmt)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.visit_AnyFunctionDef(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self.visit_AnyFunctionDef(node)

    def visit_Lambda(self, node: ast.Lambda) -> None:
        self.visit_AnyFunctionDef(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        return error.error("nested classes unsupported", node)

    # -------------------------------------------------------------------------------- #
    # Comprehensions and generators
    # -------------------------------------------------------------------------------- #

    def _visit_any_comprehension_or_generator_expr(
        self,
        node: ast.ListComp | ast.SetComp | ast.DictComp | ast.GeneratorExp,
        names: list[ast.expr],
    ) -> None:
        with new_context(self):
            # Visit the comprehensions first as they may define some of the names in
            # `names`, thus avoiding erroneous "potentially undefined" warnings.
            for comprehension in node.generators:
                self.visit(comprehension)

            for name in names:
                self.visit(name)

    def visit_comprehension(self, node: ast.comprehension) -> None:
        # Add the target (i.e. in `for x in xs` then `x` is the target) to the context
        # first to avoid "'x' potentially undefined".
        self.context.add_identifiers_to_context(node.target)

        for expr in (node.target, node.iter, *node.ifs):
            self.visit(expr)

    def visit_ListComp(self, node: ast.ListComp) -> None:
        self._visit_any_comprehension_or_generator_expr(node, [node.elt])

    def visit_SetComp(self, node: ast.SetComp) -> None:
        self._visit_any_comprehension_or_generator_expr(node, [node.elt])

    def visit_DictComp(self, node: ast.DictComp) -> None:
        self._visit_any_comprehension_or_generator_expr(node, [node.key, node.value])

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> None:
        self._visit_any_comprehension_or_generator_expr(node, [node.elt])

    # ----------------------------------------------------------------------- #
    # Special cases
    # ----------------------------------------------------------------------- #

    def visit_ReturnValue(self, node: ast.expr | None) -> bool:
        """Helper method to handle a return value / tuple elt.

        Returns True if the return value has been fully handled.

        """
        if node is None:
            return True

        if isinstance(node, (ast.Tuple, ast.List, ast.Set)):
            for elt in node.elts:
                handled = self.visit_ReturnValue(elt)

                if not handled:
                    self.visit(elt)

            return True

        if isinstance(node, ast.Dict):
            for elt in (*node.keys, *node.values):
                handled = self.visit_ReturnValue(elt)

                if not handled:
                    self.visit(elt)

            return True

        if isinstance(node, ast.Call):
            # NOTE
            #   get_basename_fullname_pair on getattr, etc. produces a name
            #   incompatible with Context::get_call_target
            if any(is_call_to(f, node) for f in PYTHON_ATTR_ACCESS_BUILTINS):
                return False

            target = self.context.get_call_target(
                fullname_of(node, safe=True),
                node,
                warn=False,
            )

            if not isinstance(target, Class):
                return False

            # Create call to class initialiser
            class_name = fullname_of(node)
            init_body = self.context.get_call_target(class_name, node)
            call = Call.from_call(
                class_name,
                call=node,
                target=init_body,
                self="@ReturnValue",
            )
            self.func_ir["calls"].add(call)

            # Visit call arguments
            for arg in (*node.args, *node.keywords):
                self.visit(arg)

            return True

        return False

    def visit_Return(self, node: ast.Return) -> None:
        """Visit ast.Return(value)."""
        handled = self.visit_ReturnValue(node.value)

        if not handled:
            self.visit(node.value)

    # ----------------------------------------------------------------------- #
    # THE FORBIDEN ZONE: A zone... that is, yes... FORBIDDEN to you.
    # ----------------------------------------------------------------------- #

    def visit_Global(self, node: ast.Global) -> None:
        return error.fatal("do not use global keyword", node)

    def visit_Nonlocal(self, node: ast.Nonlocal) -> None:
        return error.fatal("do not use nonlocal keyword", node)

    def visit_Import(self, node: ast.Import) -> None:
        return error.fatal("imports must be at the top level", node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        return error.fatal("imports must be at the top level", node)



from __future__ import annotations

import copy
from collections.abc import MutableMapping
from typing import TYPE_CHECKING

import attrs
from attrs import field

from rattr.models.context import Context
from rattr.models.ir.function import FunctionIr
from rattr.models.symbol import UserDefinedCallableSymbol

if TYPE_CHECKING:
    from collections.abc import Iterator


@attrs.mutable
class FileIr(MutableMapping[UserDefinedCallableSymbol, FunctionIr]):
    """The Intermediate Representation (IR) for the functions/classes in a file."""

    context: Context
    _file_ir: dict[UserDefinedCallableSymbol, FunctionIr] = field(
        factory=dict,
        alias="file_ir",
    )

    def ir_as_dict(self) -> dict[UserDefinedCallableSymbol, FunctionIr]:
        """Return a copy of the underlying IR dictionary."""
        return copy.deepcopy(self._file_ir)

    # ================================================================================ #
    # Mutable mapping abstract methods and mixin-overrides
    # ================================================================================ #

    def __getitem__(self, __key: UserDefinedCallableSymbol) -> FunctionIr:
        return self._file_ir.__getitem__(__key)

    def __setitem__(
        self,
        __key: UserDefinedCallableSymbol,
        __value: FunctionIr,
    ) -> None:
        return self._file_ir.__setitem__(__key, __value)

    def __delitem__(self, __key: UserDefinedCallableSymbol) -> None:
        return self._file_ir.__delitem__(__key)

    def __iter__(self) -> Iterator[UserDefinedCallableSymbol]:
        return self._file_ir.__iter__()

    def __len__(self) -> int:
        return self._file_ir.__len__()

    def clear(self) -> None:
        """Clear the File IR contents.

        ### Note
        This does not clear the associated context as that makes little sense (one may
        wish to clear the function IRs within but it does not make sense to also destroy
        the namespace).
        If you do wish to clear the context as well, you must do so directly.
        """
        # NOTE
        # The default mixin clear iterates every key and pops, this is much slower than
        # deferring to the underlying dictionary clear.
        return self._file_ir.clear()



from __future__ import annotations

import ast
from typing import TYPE_CHECKING

from rattr.ast._util import (  # noqa: F401
    get_python_attr_access_fn_obj_attr_pair,
    is_call_to_fn,
    is_string_literal,
    names_of,
)
from rattr.ast.types import AstNodeWithName

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Final

    from rattr.ast.types import Identifier
    from rattr.versioning.typing import TypeAlias

    _NameGetter: TypeAlias = Callable[[ast.expr], Identifier]


NAMEDTUPLE_INVALID_SIGNATURE_ERROR: Final = (
    "namedtuple expects exactly two positional arguments (i.e. name, attrs)"
)
NAMEDTUPLE_INVALID_SECOND_PARAMETER_VALUE_ERROR: Final = (
    "namedtuple expects the second positional argument to be a list of valid "
    "identifiers as either a Python list of strings or a string literal of space "
    "delimited strings"
)


def basename_of(
    node: ast.expr,
    *,
    unravel_attr_access_calls: bool = True,
    safe: bool = False,
) -> Identifier:
    """Return the basename of the given expression."""
    basename, _ = names_of(
        node,
        unravel_attr_access_calls=unravel_attr_access_calls,
        safe=safe,
    )
    return basename


def fullname_of(
    node: ast.expr,
    *,
    unravel_attr_access_calls: bool = True,
    safe: bool = False,
) -> Identifier:
    """Return the fullname of the given expression."""
    _, fullname = names_of(
        node,
        unravel_attr_access_calls=unravel_attr_access_calls,
        safe=safe,
    )
    return fullname


def unravel_names(
    node: ast.expr,
    *,
    _get_name: _NameGetter = basename_of,
) -> list[str]:
    """Return the name of each nameable in a given expression.

    >>> ravelled_names = ast.parse("a, b = 1, 2").body[0].targets[0]
    >>> unravel_names(ravelled_names)
    ["a", "b"]

    >>> ravelled_names = ast.parse("(a, b), c, d.e = ...").body[0].targets[0]
    >>> unravel_names(ravelled_names)
    ["a", "b", "c", "d"]

    >>> # The name getter can be overridden for example as `fullname_of`:
    >>> ravelled_names = ast.parse("a.attr = 1").body[0].targets[0]
    >>> list(unravel_names(ravelled_names))
    ["a"]
    >>> unravel_names(ravelled_names, _get_name=fullname_of)
    ["a.attr"]
    >>> # And in the complex example from above, notice "d.e" not just "d"
    >>> ravelled_names = ast.parse("(a, b), c, d.e = ...").body[0].targets[0]
    >>> unravel_names(ravelled_names, _get_name=fullname_of)
    ["a", "b", "c", "d.e"]
    """
    if isinstance(node, AstNodeWithName):
        return [_get_name(node)]

    if isinstance(node, (ast.Tuple, ast.List)):
        return [
            name
            for elt in node.elts
            for name in unravel_names(elt, _get_name=_get_name)
        ]

    raise TypeError(f"line {node.lineno}: {ast.dump(node)}")


def is_starred_import(node: ast.stmt) -> bool:
    """Return `True` if the given ast node is a starred import.

    >>> is_starred_import(ast.parse("from math import *").body[0])
    True
    >>> is_starred_import(ast.parse("from math import pi, pow").body[0])
    False
    """
    return (
        isinstance(node, ast.ImportFrom)
        and len(node.names) == 1
        and node.names[0].name == "*"
    )


def is_relative_import(node: ast.stmt) -> bool:
    """Return `True` if the given ast node is a relative import.

    >>> is_relative_import(ast.parse("from math import pi").body[0])
    False
    >>> is_relative_import(ast.parse("from .my_maths import pi").body[0])
    True
    """
    # As per documentation[1] `ImportFrom.level` is the level of the relative import,
    # where 0 means it is an absolute import.
    # [1] - https://docs.python.org/3/library/ast.html#ast.ImportFrom
    return isinstance(node, ast.ImportFrom) and node.level != 0


def assignment_targets(
    assignment: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> list[ast.expr]:
    """Return the assignment targets (assignment lhs)."""
    if isinstance(assignment, ast.Assign):
        return assignment.targets
    if isinstance(assignment, (ast.AnnAssign, ast.AugAssign, ast.NamedExpr)):
        return [assignment.target]
    else:
        raise TypeError(f"line {assignment.lineno}: {ast.dump(assignment)}")


def assignment_is_one_to_one(
    assignment: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> bool:
    """Return `True` if the given assignment is one-to-one."""
    targets = assignment_targets(assignment)

    def _is_iterable(target: ast.expr | None) -> bool:
        return isinstance(target, (ast.Tuple, ast.List))

    lhs_is_singular = len(targets) == 1 and not _is_iterable(targets[0])
    rhs_is_singular = not _is_iterable(assignment.value)

    return lhs_is_singular and rhs_is_singular


def has_lambda_in_rhs(
    assignment: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> bool:
    """Return `True` if the given assignment has a lambda in the right-hand side."""
    if isinstance(assignment.value, ast.Lambda):
        return True
    elif isinstance(assignment.value, (ast.Tuple, ast.List)):  # iterable unpacking
        return any(isinstance(v, ast.Lambda) for v in assignment.value.elts)
    else:
        return False


def has_walrus_in_rhs(
    assignment: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> bool:
    """Return `True` if the given assignment contains a walrus assignment in the rhs."""
    if isinstance(assignment.value, ast.NamedExpr):
        return True
    elif isinstance(assignment.value, (ast.Tuple, ast.List)):
        return any(isinstance(v, ast.NamedExpr) for v in assignment.value.elts)
    else:
        return False


def walruses_in_rhs(
    assignment: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> list[ast.NamedExpr]:
    """Return the walruses in the rhs of the given assignment.

    >>> walruses_in_rhs(ast.parse("a = (b := c)").body[0].value)
    [NamedExpr(target=Name(id='b', ctx=Store()), value=Name(id='c', ctx=Load())), ]

    >>> walruses_in_rhs(ast.parse("a = (b, c := d)").body[0].value)
    [NamedExpr(target=Name(id='c', ctx=Store()), value=Name(id='d', ctx=Load())), ]
    """
    if not has_walrus_in_rhs(assignment):
        return []

    if isinstance(assignment.value, (ast.Tuple, ast.List)):
        values = assignment.value.elts
    else:
        values = [assignment.value]

    return [e for e in values if isinstance(e, ast.NamedExpr)]


def has_namedtuple_declaration_in_rhs(
    assignment: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> bool:
    """Return `True` if the rhs contains a call to create a namedtuple type."""

    def _target_is_namedtuple(call: ast.Call) -> bool:
        # HACK
        # Naive approach, replace when enum / class namedtuple / etc is improved.
        # See: cls.py::is_enum, cls.py::is_namedtuple
        name = fullname_of(call.func, safe=True)
        return name == "namedtuple" or name.endswith(".namedtuple")

    if isinstance(assignment.value, ast.Call):
        return _target_is_namedtuple(assignment.value)
    elif isinstance(assignment.value, (ast.Tuple, ast.List)):
        return any(
            _target_is_namedtuple(value)
            for value in assignment.value.elts
            if isinstance(value, ast.Call)
        )
    else:
        return False


def namedtuple_init_signature_from_declaration(
    assignment: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
) -> tuple[list[str], dict[str, str]]:
    """Return the args/attrs of the namedtuple constructed by this assignment by call.

    Note:
        Assumes one-to-one assignment.

    >>> node = ast.parse("p = namedtuple('p', ['x', 'y'])").body[0].value
    >>> namedtuple_init_signature_from_declaration(node)
    ["self", "x", "y]

    Raises:
        TypeError: The node is not an assignment of a call.
        ValueError: Invalid namedtuple declaration.

    Returns:
        tuple[list[str], dict[str, str]]: The positional and keyword args of the init.
    """

    def namedtuple_attrs_from_second_argument(attrs: ast.expr) -> list[str]:
        if isinstance(attrs, ast.List):
            return unpack_ast_list_of_strings(attrs)

        if isinstance(attrs, ast.Constant):
            return parse_space_delimited_ast_string(attrs)

        raise SyntaxError

    # Parse call arguments
    if not isinstance(assignment.value, ast.Call):
        raise TypeError

    namedtuple_call_arguments = assignment.value.args

    if len(namedtuple_call_arguments) != 2:
        raise ValueError(NAMEDTUPLE_INVALID_SIGNATURE_ERROR)

    _, namedtuple_attrs_argument = namedtuple_call_arguments

    try:
        attrs = namedtuple_attrs_from_second_argument(namedtuple_attrs_argument)
    except SyntaxError as exc:
        raise ValueError(NAMEDTUPLE_INVALID_SECOND_PARAMETER_VALUE_ERROR) from exc

    return ["self", *attrs]


def unpack_ast_list_of_strings(ast_attrs: ast.List) -> list[str]:
    """Return the given ast.List of ast.Constant strings as a Python list[str]."""
    unpacked_attrs: list[str] = [
        arg.value
        for arg in ast_attrs.elts
        if isinstance(arg, ast.Constant)
        if isinstance(arg.value, str)
    ]

    if len(unpacked_attrs) != len(ast_attrs.elts):
        raise SyntaxError

    return unpacked_attrs


def parse_space_delimited_ast_string(ast_string: ast.Constant) -> list[str]:
    """Return the given ast.Constant string as a Python str."""
    if not isinstance(ast_string.value, str):
        raise SyntaxError

    if ast_string.value == "":
        return []

    parsed_attrs = ast_string.value.split(" ")

    if not all(attr.isidentifier() for attr in parsed_attrs):
        raise SyntaxError

    return parsed_attrs



"""Rattr Base class tests."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from rattr.analyser.base import Assertor, CustomFunctionAnalyser
from rattr.models.symbol import Call, CallArguments, Name

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tests.shared import ArgumentsFn, StateFn


@pytest.fixture(autouse=True)
def __set_current_file(state: StateFn) -> Iterator[None]:
    with state(current_file=Path(__file__)):
        yield


class TestAssertor:
    def test_assertor_is_strict(self, capfd: pytest.CaptureFixture[str]):
        assertor = Assertor(is_strict=True)

        with mock.patch("sys.exit") as _exit:
            assertor.failed("the reason")

        _, stderr = capfd.readouterr()

        assert "the reason" in stderr
        assert _exit.call_count == 1

    def test_assertor_not_is_strict(
        self,
        capfd: pytest.CaptureFixture[str],
        arguments: ArgumentsFn,
    ):
        assertor = Assertor(is_strict=False)

        with arguments(_warning_level="all"):
            with mock.patch("sys.exit") as _exit:
                assertor.failed("the reason2")

        _, stderr = capfd.readouterr()

        assert "the reason" in stderr
        assert _exit.call_count == 0


def test_custom_function_analyser_on_def(
    builtin_print_analyser: CustomFunctionAnalyser,
    example_func_analyser: CustomFunctionAnalyser,
):
    mock_call = mock.Mock()
    mock_context = mock.Mock()

    assert builtin_print_analyser.on_def("print", mock_call, mock_context) == {
        "gets": {Name("get_in_print_def")},
        "sets": {Name("set_in_print_def")},
        "dels": {Name("del_in_print_def")},
        "calls": {
            Call(
                name="call_in_print_def",
                args=CallArguments(args=(), kwargs={}),
                target=None,
            ),
        },
    }

    assert example_func_analyser.on_def("example", mock_call, mock_context) == {
        "gets": {Name("get_in_example_def")},
        "sets": {Name("set_in_example_def")},
        "dels": {Name("del_in_example_def")},
        "calls": {
            Call(
                name="call_in_example_def",
                args=CallArguments(args=(), kwargs={}),
                target=None,
            ),
        },
    }


def test_custom_function_analyser_on_call(
    builtin_print_analyser: CustomFunctionAnalyser,
    example_func_analyser: CustomFunctionAnalyser,
):
    mock_call = mock.Mock()
    mock_context = mock.Mock()

    assert builtin_print_analyser.on_call("print", mock_call, mock_context) == {
        "gets": {Name("get_in_print")},
        "sets": {Name("set_in_print")},
        "dels": {Name("del_in_print")},
        "calls": {
            Call(
                name="call_in_print",
                args=CallArguments(args=(), kwargs={}),
                target=None,
            ),
        },
    }

    assert example_func_analyser.on_call("example", mock_call, mock_context) == {
        "gets": {Name("get_in_example")},
        "sets": {Name("set_in_example")},
        "dels": {Name("del_in_example")},
        "calls": {
            Call(
                name="call_in_example",
                args=CallArguments(args=(), kwargs={}),
                target=None,
            ),
        },
    }



from __future__ import annotations

from enum import Enum


class TomlArgumentType(Enum):
    flag = "bool"
    """I.e. a boolean flag.

    In toml:
        variable=true

    In sys argument list:
        --variable
    """

    int = "int"
    """I.e. a single int value.

    In toml:
        variable=42

    In sys argument list:
        --variable 42
    """

    string = "str"
    """I.e. a single string value.

    In toml:
        variable="some value"

    In sys argument list:
        --variable "some value"
    """

    list_of_strings = "list[str]"
    """I.e. a list of values.

    In toml:
        variable=["value", "another value", "etc"]

    In sys argument list:
        --variable "value" --variable "another value" --variable "etc"
    """

    def __str__(self) -> str:
        return self.value

    def is_valid(self, value: bool | int | str | list[str]) -> bool:
        if self == TomlArgumentType.flag:
            return isinstance(value, bool)

        if self == TomlArgumentType.int:
            return isinstance(value, int)

        if self == TomlArgumentType.string:
            return isinstance(value, str)

        if self == TomlArgumentType.list_of_strings:
            return isinstance(value, list) and all(isinstance(_v, str) for _v in value)

        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest

import tests.helpers as helpers
from rattr.models.context import Context
from rattr.plugins.assertors.import_clobbering import ImportClobberingAssertor

if TYPE_CHECKING:
    import ast
    from typing import Callable


class TestImportNameClobbering:
    def test_would_be_clobbered_but_is_renamed(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi as pie

            pi = 3.2 - 0.06 + 0.001 # ...
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert not _exit.called

    def test_clobber_at_module_level(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            import math

            math = "this is very bad"
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'math'")],
        )

        assert _exit.called

    def test_clobber_in_child_context(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from os.path import join

            def func():
                join = "this" + "joined with" + "this"
                return join
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'join'")],
        )

        assert _exit.called

    def test_clobber_in_multiple_assignment(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from os.path import join

            def func():
                a, b, join, d = 1, 2, 3, 4
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'join'")],
        )

        assert _exit.called

    def test_clobber_in_annotated_assignment(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            import math

            math: str = "this is very bad"
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'math'")],
        )

        assert _exit.called

    def test_clobber_in_augmented_assignment(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            pi += 1.0   # that's not pi!
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_function_definition(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from os.path import join

            def join(*args):
                return "/".join(args)
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'join'")],
        )

        assert _exit.called

    def test_clobber_by_nested_function_definition(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from os.path import join

            def wrapper():
                def join(*args):
                    return "/".join(args)
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'join'")],
        )

        assert _exit.called

    def test_clobber_by_async_function_definition(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from os.path import join

            async def join(*args):
                return "/".join(args)
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'join'")],
        )

        assert _exit.called

    def test_clobber_by_class_definition(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from pathlib import Path

            class Path:
                ...
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'Path'")],
        )

        assert _exit.called

    def test_clobber_by_argument_name(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            def area(r, pi):
                return pi * r * r
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_argument_name_in_nested_function(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            def wrapped():
                def area(r, pi):
                    return pi * r * r
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_argument_name_in_async_function(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            async def area(r, pi):
                return pi * r * r
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_argument_name_in_lambda(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            x = lambda pi: 3.14
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_iterator_name_in_for_loop(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            for pi in pies:
                eat(pi)  # yum
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_iterator_name_in_for_loop_while_unpacking(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            for a, pi in pies:
                eat(pi)
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_iterator_name_in_async_for_loop(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            async for pi in pies:
                eat(pi)  # yum
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_name_in_with_statement(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            def fn():
                with whatever as pi:
                    pass
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_name_in_with_statement_while_unpacking(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            def fn():
                with whatever as we, whenever as pi:
                    pass
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_name_in_async_with_statement(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            def fn():
                async with whatever as pi:
                    pass
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_iterator_name_in_list_comprehension(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            l = [pi for pi in pies]
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_iterator_name_in_set_comprehension(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            s = {pi for _, pi in pastries_and_pies}
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_iterator_name_in_dict_comprehension(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            r"""
            from math import pi

            d = {pi: filling for pi, filling in pies}
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_clobber_by_iterator_name_in_generator_expression(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            is_any = any(pi for pi in pies)
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called


class TestDoNotFailOnPotentialFalsePositives:
    def test_allow_attribute_write(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
    ):
        _ast, _ctx = parse_with_context(
            """
            from pathlib import Path
            from math import pi

            Path.attr = "on class"
            MyClass.pi = "on attr"
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert not _exit.called

    def test_allow_attribute_write_in_chained_assignment(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
    ):
        _ast, _ctx = parse_with_context(
            """
            from pathlib import Path
            from math import pi

            thing = Path.attr = "on class"
            thing = MyClass.pi = "on attr"
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert not _exit.called


class TestMultipleClobbersWithLocation:
    def test_multiple_failures_with_location(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            def fn():
                del pi

                def inner():
                    pi = 1

                del pi, again
                pi = 2
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [
                helpers.as_fatal("attempt to delete imported name 'pi'", line=4, col=4),
                helpers.as_fatal("redefinition of imported name 'pi'", line=7, col=8),
                helpers.as_fatal("attempt to delete imported name 'pi'", line=9, col=4),
                helpers.as_fatal("redefinition of imported name 'pi'", line=10, col=4),
            ],
        )

        assert _exit.called


class TestIgnoreAnnotatedClobbers:
    def test_ignore_clobbers_with_rattr_ignore(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import func, async_func, SomeClass

            @rattr_ignore()
            def func():
                pass

            @rattr_ignore()
            async def async_func():
                pass

            @rattr_ignore()
            class SomeClass:
                pass
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert not _exit.called

    def test_ignore_clobbers_with_rattr_results(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import func, async_func, SomeClass

            @rattr_results()
            def func():
                pass

            @rattr_results()
            async def async_func():
                pass

            @rattr_results()
            class SomeClass:
                pass
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert not _exit.called


class TestDeleteImportedName:
    def test_delete_imported_name(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            def fn():
                del pi
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("attempt to delete imported name 'pi'")],
        )

        assert _exit.called

    def test_delete_imported_name_in_multiple_delete(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            def fn():
                del a, b, pi
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("attempt to delete imported name 'pi'")],
        )

        assert _exit.called


class TestNameClobberedByMethod:
    def test_simple_failure(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
        capfd: pytest.CaptureFixture[str],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            pizza = 5  # good
            pi = 1  # bad, the name `bar` is clobbered
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert helpers.stderr_matches(
            capfd,
            [helpers.as_fatal("redefinition of imported name 'pi'")],
        )

        assert _exit.called

    def test_do_not_fail_on_method_name(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            class MyClass:
                def pi(self):
                    ...
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert not _exit.called

    def test_do_not_fail_on_property_name(
        self,
        parse_with_context: Callable[[str], tuple[ast.AST, Context]],
    ):
        _ast, _ctx = parse_with_context(
            """
            from math import pi

            class MyClass:
                @property
                def pi(self):
                    ...
            """
        )

        with mock.patch("sys.exit") as _exit:
            ImportClobberingAssertor(is_strict=True).assert_holds(_ast, _ctx)

        assert not _exit.called



"""Rattr error/logging functions."""
from __future__ import annotations

import ast
import sys
from enum import Enum
from typing import NoReturn

from rattr.config import Config, ShowWarnings
from rattr.models.symbol import Symbol

__ERROR = "{prefix}: {optional_file_info}{optional_line_info}: {message}"
__FILE_INFO = "\033[1m{}\033[0m"
__LINE_INFO = "\033[1m:{}:{}\033[0m"


# --------------------------------------------------------------------------- #
# Rattr errors
# --------------------------------------------------------------------------- #


class Level(Enum):
    rattr = "\033[34;1mrattr\033[0m"  # Blue
    info = "\033[33;1minfo\033[0m"  # Yellow / Orange
    warning = "\033[33;1mwarning\033[0m"  # Yellow / Orange
    error = "\033[31;1merror\033[0m"  # Red
    fatal = "\033[31;1mfatal\033[0m"  # Red


def rattr(
    message: str,
    culprit: ast.AST | Symbol | None = None,
    badness: int = 0,
) -> None:
    """Log a message with the prefix "rattr", not for analyser errors."""
    __log(Level.rattr, message, culprit)


def info(
    message: str,
    culprit: ast.AST | Symbol | None = None,
    badness: int = 0,
) -> None:
    """Log a low-priority warning and, if given, include culprit info."""
    config = Config()
    config.increment_badness(badness)

    if config.do_not_show_warnings:
        return

    if config.is_in_target_file:
        warning_level = ShowWarnings.target_low_priority
    else:
        warning_level = ShowWarnings.inherited_low_priority

    if warning_level not in config.arguments.show_warnings:
        return

    __log(Level.info, message, culprit)


def warning(
    message: str,
    culprit: ast.AST | Symbol | None = None,
    badness: int = 1,
) -> None:
    """Log a warning and, if given, include culprit line and file info."""
    config = Config()
    config.increment_badness(badness)

    if config.do_not_show_warnings:
        return

    if config.is_in_target_file:
        warning_level = ShowWarnings.target
    else:
        warning_level = ShowWarnings.inherited_high_priority

    if warning_level not in config.arguments.show_warnings:
        return

    __log(Level.warning, message, culprit)


def error(
    message: str,
    culprit: ast.AST | Symbol | None = None,
    badness: int = 5,
) -> None:
    """Log an error and, if given, include culprit line and file info."""
    config = Config()
    config.increment_badness(badness)

    if badness > 0 and config.arguments.is_strict:
        fatal(message, culprit)

    __log(Level.error, message, culprit)


def fatal(
    message: str,
    culprit: ast.AST | Symbol | None = None,
    badness: int = 0,  # noqa
) -> NoReturn:
    """Log a fatal error and, if given, include culprit line and file info.

    NOTE
        A fatal error has no notion of badness as it will always cause an
        immediate EXIT_FAILURE, however, badness is provided in the function
        interface for consistency with the other errors.

        Regardless of the provided badness value, a badness of 0 will be used.

    """
    config = Config()
    config.increment_badness(badness)

    __log(Level.fatal, message, culprit)

    sys.exit(1)


def get_file_and_line_info(culprit: ast.AST | Symbol | None) -> tuple[str, str]:
    """Return the formatted line and line and file info as strings."""
    return __file_info(culprit), __line_info(culprit)


def __file_info(culprit: ast.AST | Symbol | None) -> str:
    config = Config()

    current_file = config.formatted_current_file_path

    if isinstance(culprit, Symbol) and (location := culprit.location) is not None:
        file = config.get_formatted_path(location.file)
    elif current_file:
        file = current_file
    else:
        file = config.formatted_target_path

    if file is None:
        file = ""  # never

    return __FILE_INFO.format(file)


def __line_info(culprit: ast.AST | Symbol | None) -> str:
    if isinstance(culprit, Symbol):
        if (location := culprit.location) is not None:
            line_info = __LINE_INFO.format(location.lineno, location.col_offset)
        else:
            line_info = ""
    elif culprit is None:
        line_info = ""
    else:
        line_info = __LINE_INFO.format(culprit.lineno, culprit.col_offset)

    return line_info


def __log(
    level: Level,
    message: str,
    culprit: ast.AST | Symbol | None = None,
) -> None:
    file_info, line_info = get_file_and_line_info(culprit)

    print(
        __ERROR.format(
            prefix=level.value,
            optional_file_info=file_info,
            optional_line_info=line_info,
            message=message,
        ),
        file=sys.stderr,
    )



"""Rattr class analyser.

Rattr is designed to be a function analyser and thus the class-based features
are rather limited.

"""
from __future__ import annotations

import ast
from typing import TYPE_CHECKING

from rattr.analyser.base import NodeVisitor
from rattr.analyser.function import FunctionAnalyser
from rattr.analyser.types import ClassIr
from rattr.analyser.util import has_annotation, parse_rattr_results_from_annotation
from rattr.ast.util import assignment_targets, fullname_of, unravel_names
from rattr.models.context import Context
from rattr.models.symbol import CallInterface, Class, Func, Name

if TYPE_CHECKING:
    from collections.abc import Iterator

    from rattr.ast.types import Identifier


def is_method(node: ast.stmt) -> bool:
    return isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))


def init_method_or_none(
    methods: list[ast.FunctionDef | ast.AsyncFunctionDef],
) -> ast.FunctionDef | None:
    init_methods = [method for method in methods if method.name == "__init__"]

    if len(init_methods) == 0:
        return None

    return init_methods[0]


def iter_static_methods(
    methods: list[ast.FunctionDef | ast.AsyncFunctionDef],
) -> Iterator[ast.FunctionDef | ast.AsyncFunctionDef]:
    return (method for method in methods if has_annotation("staticmethod", method))


def base_names(cls: ast.ClassDef) -> list[Identifier]:
    return [fullname_of(b, safe=True) for b in cls.bases]


def is_enum_by_heuristic(cls: ast.ClassDef) -> bool:
    return any(b == "Enum" or b.endswith(".Enum") for b in base_names(cls))


def is_namedtuple_by_heuristic(cls: ast.ClassDef) -> bool:
    return any(b == "NamedTuple" or b.endswith(".NamedTuple") for b in base_names(cls))


class ClassAnalyser(NodeVisitor):
    """Walk a class's AST to determine the accessed attributes.


    TODO
        - Inheritance, super
        - Class methods
        - Methods from type
        - Nested classes

    """

    def __init__(self, _ast: ast.ClassDef, context: Context) -> None:
        """Set configuration and initialise IR."""
        if not isinstance(_ast, ast.ClassDef):
            raise TypeError("ClassAnalyser expects `_ast` to be a class")

        self._ast = _ast
        self.name = _ast.name

        self.class_ir: ClassIr = {}

        self.context = context

    def analyse(self) -> ClassIr:
        """Entry point, return the IR produced from analysis.

        Algorithm:
            * Analyse non-method body -- i.e. class attributes, etc
                Add to context as `ClassName.attr`
                TODO Allow reference from class body context as just `attr`
            * Analyse __init__
                Add to context as Func: `ClassName`
                TODO deal with __new__, __new__ and __init__, etc
            * Analyse `@static_methods`
                Add to context as Func: `ClassName.method_name`
            * TODO Analyse `@class_methods`
            * TODO Analyse remaining methods as object methods

        """

        # NOTE
        #   Analyse within parent context, i.e. no new_context, as static
        #   methods, initialisers, should be present in the parent context
        #   just with transformed names

        statements = [stmt for stmt in self._ast.body if not is_method(stmt)]
        methods = [stmt for stmt in self._ast.body if is_method(stmt)]

        for stmt in statements:
            self.visit(stmt)

        init = init_method_or_none(methods)

        if init is not None:
            self.visit_initialiser(init)

        # HACK Default initialisers for special cases
        if init is None:
            if is_enum_by_heuristic(self._ast):
                self.visit_enum_initialiser()

            if is_namedtuple_by_heuristic(self._ast):
                self.visit_named_tuple_initialiser()

        for method in iter_static_methods(methods):
            self.visit_static_method(method)

        return self.class_ir

    # ----------------------------------------------------------------------- #
    # Special class helpers
    # ----------------------------------------------------------------------- #

    def visit_initialiser(self, init: ast.FunctionDef) -> None:
        if has_annotation("rattr_ignore", self._ast):
            return

        new_symbol = self.symbol.with_init(init)
        self.update_symbol(new_symbol)

        if has_annotation("rattr_results", self._ast):
            self.class_ir[new_symbol] = parse_rattr_results_from_annotation(
                self._ast,
                context=self.context,
            )
            return

        init_analyser = FunctionAnalyser(init, self.context)
        self.class_ir[new_symbol] = init_analyser.analyse()

    def visit_enum_initialiser(self) -> None:
        new_symbol = self.symbol.with_init_arguments(args=["self", "_id"])
        self.update_symbol(new_symbol)

        self.class_ir[new_symbol] = {
            "sets": set(),
            "gets": {
                symbol
                for symbol in self.context.symbol_table.symbols
                if isinstance(symbol, Name)
                if symbol.name.startswith(self.prefix)
            },
            "calls": set(),
            "dels": set(),
        }

    def visit_named_tuple_initialiser(self) -> None:
        items = [
            symbol.name[len(self.prefix) :]
            for symbol in self.context.symbol_table.symbols
            if isinstance(symbol, Name)
            if symbol.name.startswith(self.prefix)
        ]

        cls = self.update_symbol(self.symbol.with_init_arguments(args=["self", *items]))
        self.class_ir[cls] = {
            "sets": set(),
            "gets": set(),
            "calls": set(),
            "dels": set(),
        }

    def visit_static_method(
        self,
        method: ast.FunctionDef | ast.AsyncFunctionDef,
    ) -> None:
        qualified_name = f"{self.name}.{method.name}"

        fn = Func(
            name=qualified_name,
            token=method,
            interface=CallInterface.from_fn_def(method),
        )
        self.context.add(fn)

        method_analyser = FunctionAnalyser(method, self.context)
        self.class_ir[fn] = method_analyser.analyse()

    # ----------------------------------------------------------------------- #
    # Non-method statements
    # ----------------------------------------------------------------------- #

    def visit_AnyAssign(
        self,
        node: ast.Assign | ast.AnnAssign | ast.AugAssign | ast.NamedExpr,
    ) -> None:
        """Helper method for assignments.

        Visit ast.Assign(targets, value, type_comment)\\
        Visit ast.AnnAssign(target, annotation, value, simple)\\
        Visit ast.AugAssign(target, op, value)

        """
        names = [
            name
            for target in assignment_targets(node)
            for name in unravel_names(target)
        ]

        for name in names:
            self.context.add(Name(f"{self.name}.{name}", self.name, token=node))

        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        self.visit_AnyAssign(node)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        self.visit_AnyAssign(node)

    def visit_NamedExpr(self, node: ast.NamedExpr) -> None:
        self.visit_AnyAssign(node)

    # ----------------------------------------------------------------------- #
    # Helpers
    # ----------------------------------------------------------------------- #

    @property
    def prefix(self) -> str:
        return f"{self.name}."

    @property
    def symbol(self) -> Class:
        cls = self.context.get(self.name)

        if not isinstance(cls, Class):
            raise ValueError(f"class {self.name} is not in the current context")

        return cls

    def update_symbol(self, new_class_symbol: Class) -> Class:
        _ = self.context.pop(new_class_symbol.id)
        self.context[new_class_symbol.id] = new_class_symbol
        return new_class_symbol



from __future__ import annotations

import ast
import json
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from frozendict import frozendict

from rattr.models.context import Context, SymbolTable, compile_root_context
from rattr.models.ir import FileIr, FunctionIr
from rattr.models.results import FileResults, FunctionResults
from rattr.models.symbol import (
    AnyCallInterface,
    Builtin,
    Call,
    CallArguments,
    CallInterface,
    Class,
    Func,
    Import,
    Location,
    Name,
    Symbol,
    UserDefinedCallableSymbol,
)
from rattr.models.util import OutputIrs, deserialise, serialise, serialise_irs
from tests.models.util.shared import CallInterfaceArgs

if TYPE_CHECKING:
    from tests.shared import (
        MakeRootContextFn,
        MakeSymbolTableFn,
        OsDependentPathFn,
        ParseFn,
    )


@pytest.fixture
def symbol_group_one() -> list[Symbol]:
    return [
        Name("bob"),
        Func("my_func", interface=CallInterface(posonlyargs=("a", "b"))),
        Name("foo.bar.baz", "foo"),
    ]


@pytest.fixture
def symbol_group_two() -> list[Symbol]:
    return [
        Name("i.am.in.the.child.context", "i"),
        Func(
            "a_function_here",
            interface=CallInterface(args=("x", "y"), vararg="extras"),
        ),
    ]


@pytest.fixture
def simple_context_one(
    make_root_context: MakeRootContextFn,
    symbol_group_one: list[Symbol],
) -> Context:
    return make_root_context(symbol_group_one, include_root_symbols=False)


@pytest.fixture
def simple_context_two(
    make_root_context: MakeRootContextFn,
    symbol_group_two: list[Symbol],
) -> Context:
    return make_root_context(symbol_group_two, include_root_symbols=False)


@pytest.fixture
def nested_context(
    make_symbol_table: MakeSymbolTableFn,
    simple_context_one: Context,
    symbol_group_two: list[Symbol],
):
    return Context(
        parent=simple_context_one,
        symbol_table=make_symbol_table(symbol_group_two, include_root_symbols=False),
        file=Path("/some/other/file.py"),
    )


@pytest.fixture
def full_file_ir() -> dict[UserDefinedCallableSymbol, FunctionIr]:
    my_func_symbol = Func("my_func", interface=CallInterface(posonlyargs=("a", "b")))
    a_function_here_symbol = Func(
        "a_function_here",
        interface=CallInterface(args=("x", "y"), vararg="extras"),
    )
    my_class_symbol = Class(
        "MyClass",
        interface=CallInterface(
            args=("self", "data"),
            kwonlyargs=("kwarg_a", "kwarg_b"),
        ),
    )

    return {
        my_func_symbol: FunctionIr.new(
            gets={
                Name("a"),
                Name("b.bar.baz", "b"),
            }
        ),
        a_function_here_symbol: FunctionIr.new(
            gets={
                Name("foo"),
                Name("foo.bar.baz", "foo"),
            },
            calls={
                Call(
                    name=my_func_symbol.name,
                    args=CallArguments(args=("foo", "foo")),
                    target=my_func_symbol,
                ),
            },
        ),
        my_class_symbol: FunctionIr.new(
            gets={
                Name("self"),
                Name("data"),
                Name("kwarg_a"),
                Name("kwarg_b"),
            },
            sets={
                Name("self.data", "self"),
                Name("self._a", "self"),
                Name("self._b", "self"),
            },
        ),
    }


def test_name(parse: ParseFn, os_dependent_path: OsDependentPathFn):
    ast_module = parse(
        """
        def _():
            x = "bloop"
        """
    )
    ast_fn_def: ast.FunctionDef = ast_module.body[0]
    ast_assign: ast.Assign = ast_fn_def.body[0]
    ast_name: ast.Name = ast_assign.targets[0]

    name = Name("x", token=ast_name)

    serialised = serialise(name)
    assert json.loads(serialised) == {
        "type": "Name",
        "name": "x",
        "basename": "x",
        "location": {
            "lineno": 2,
            "end_lineno": 2,
            "col_offset": 4,
            "end_col_offset": 5,
            "file": os_dependent_path("test.py"),
        },
        "interface": None,
    }

    deserialised = deserialise(serialised, type=Symbol)
    assert deserialised == name


@pytest.mark.parametrize("builtin_name", ["print", "getattr"])
def test_builtin(builtin_name: str):
    builtin = Builtin(builtin_name)

    serialised = serialise(builtin)
    assert json.loads(serialised) == {
        "type": "Builtin",
        "name": builtin_name,
        "location": {
            "lineno": 1,
            "col_offset": 0,
            "end_lineno": None,
            "end_col_offset": None,
            "file": "built-in",
        },
        "interface": "any",
    }

    deserialised = deserialise(serialised, type=Builtin)
    assert deserialised == builtin


def test_import(parse: ParseFn, os_dependent_path: OsDependentPathFn):
    ast_module = parse(
        """
        from math import pi
        """
    )
    context = compile_root_context(ast_module)
    import_: Import = context.get("pi")

    serialised = serialise(import_)
    assert json.loads(serialised) == {
        "type": "Import",
        "name": "pi",
        "qualified_name": "math.pi",
        "location": {
            "lineno": 1,
            "col_offset": 0,
            "end_lineno": 1,
            "end_col_offset": 19,
            "file": os_dependent_path("test.py"),
        },
        "interface": "any",
    }

    deserialised = deserialise(serialised, type=Symbol)
    assert deserialised == import_


def test_import_multiple(parse: ParseFn, os_dependent_path: OsDependentPathFn):
    ast_module = parse(
        """
        '''Blah.'''
        import math as maths, os
        """
    )
    context = compile_root_context(ast_module)
    imports_: list[Import] = [
        context.get("maths"),
        context.get("os"),
    ]

    serialised = [serialise(import_) for import_ in imports_]
    assert [json.loads(s) for s in serialised] == [
        {
            "type": "Import",
            "name": "maths",
            "qualified_name": "math",
            "location": {
                "lineno": 2,
                "col_offset": 0,
                "end_lineno": 2,
                "end_col_offset": 24,
                "file": os_dependent_path("test.py"),
            },
            "interface": "any",
        },
        {
            "type": "Import",
            "name": "os",
            "qualified_name": "os",
            "location": {
                "lineno": 2,
                "col_offset": 0,
                "end_lineno": 2,
                "end_col_offset": 24,
                "file": os_dependent_path("test.py"),
            },
            "interface": "any",
        },
    ]

    deserialised = [deserialise(s, type=Symbol) for s in serialised]
    assert deserialised == imports_


def test_func(parse: ParseFn, os_dependent_path: OsDependentPathFn):
    ast_module = parse(
        """
        def func(a, /, b, c, *args, d = None, e = None, **kwargs):
            ...
        """
    )
    context = compile_root_context(ast_module)
    func: Func = context.get("func")

    serialised = serialise(func)
    assert json.loads(serialised) == {
        "type": "Func",
        "name": "func",
        "location": {
            "lineno": 1,
            "col_offset": 0,
            "end_lineno": 2,
            "end_col_offset": 7,
            "file": os_dependent_path("test.py"),
        },
        "interface": {
            "posonlyargs": ["a"],
            "args": ["b", "c"],
            "vararg": "args",
            "kwonlyargs": ["d", "e"],
            "kwarg": "kwargs",
        },
        "is_async": False,
    }

    deserialised = deserialise(serialised, type=Symbol)
    assert deserialised == func


def test_class(parse: ParseFn, os_dependent_path: OsDependentPathFn):
    ast_module = parse(
        """
        class ParentClass:
            ...

        class MyClass(ParentClass):
            ...
        """
    )
    context = compile_root_context(ast_module)
    cls: Class = context.get("MyClass")

    serialised = serialise(cls)
    assert json.loads(serialised) == {
        "type": "Class",
        "name": "MyClass",
        "location": {
            "lineno": 4,
            "col_offset": 0,
            "end_lineno": 5,
            "end_col_offset": 7,
            "file": os_dependent_path("test.py"),
        },
        "interface": "any",
    }

    deserialised = deserialise(serialised, type=Symbol)
    assert deserialised == cls


def test_call(parse: ParseFn, os_dependent_path: OsDependentPathFn):
    ast_module = parse(
        """
        def some_func(a: int, b: int, **kwargs):
            ...

        y = ...
        x = some_func(y.some_attr, y.nested.attr)
        """
    )
    context = compile_root_context(ast_module)
    func: Func = context.get("some_func")
    call = Call(
        name="some_func",
        args=CallArguments(args=("y.some_attr", "y.nested.attr")),
        target=func,
    )

    serialised = serialise(call)
    assert json.loads(serialised) == {
        "type": "Call",
        "name": "some_func",
        "args": {"args": ["y.some_attr", "y.nested.attr"], "kwargs": {}},
        "target": {
            "type": "Func",
            "name": "some_func",
            "location": {
                "lineno": 1,
                "col_offset": 0,
                "end_lineno": 2,
                "end_col_offset": 7,
                "file": os_dependent_path("test.py"),
            },
            "interface": {
                "posonlyargs": [],
                "args": ["a", "b"],
                "vararg": None,
                "kwonlyargs": [],
                "kwarg": "kwargs",
            },
            "is_async": False,
        },
        "location": {
            "lineno": 1,
            "col_offset": 0,
            "end_lineno": None,
            "end_col_offset": None,
            "file": os_dependent_path("test.py"),
        },
    }

    deserialised = deserialise(serialised, type=Symbol)
    assert deserialised == call


@pytest.mark.parametrize(
    "as_tuple",
    testcases := [
        # the empty signature
        CallInterfaceArgs([], [], None, [], None),
        # just args and kwargs
        CallInterfaceArgs([], [], "args", [], None),
        CallInterfaceArgs([], [], None, [], "kwargs"),
        CallInterfaceArgs([], [], "args", [], "kwargs"),
        # posonlyargs
        CallInterfaceArgs(["a"], [], None, [], None),
        CallInterfaceArgs(["a", "b"], [], None, [], None),
        CallInterfaceArgs(["a"], [], "args", [], None),
        # args
        CallInterfaceArgs([], ["a"], None, [], None),
        CallInterfaceArgs([], ["a", "b"], None, [], None),
        CallInterfaceArgs([], ["a"], "args", [], None),
        # kwonlyargs
        CallInterfaceArgs([], [], None, ["a"], None),
        CallInterfaceArgs([], [], None, ["a", "b"], None),
        CallInterfaceArgs([], [], "args", ["a"], None),
        # mixed
        CallInterfaceArgs(["a", "b"], ["c", "d", "e"], "args", ["f", "g"], "kwargs"),
    ],
    ids=[i for i, _ in enumerate(testcases)],
)
def test_call_interface(as_tuple: CallInterfaceArgs):
    interface = CallInterface(*as_tuple)

    serialised = serialise(interface)
    assert json.loads(serialised) == {
        "posonlyargs": as_tuple.posonlyargs,
        "args": as_tuple.args,
        "vararg": as_tuple.vararg,
        "kwonlyargs": as_tuple.kwonlyargs,
        "kwarg": as_tuple.kwarg,
    }

    deserialised = deserialise(serialised, type=CallInterface)
    assert deserialised == interface

    # When serialise the tuple becomes a list, so ensure that on deserialisation this is
    # reversed
    assert isinstance(deserialised.posonlyargs, tuple)
    assert isinstance(deserialised.args, tuple)
    assert isinstance(deserialised.kwonlyargs, tuple)


def test_any_call_interface():
    interface = AnyCallInterface()

    serialised = serialise(interface)
    assert serialised == '"any"'

    deserialised = deserialise(serialised, type=AnyCallInterface)
    assert deserialised == interface


def test_any_call_interface_as_call_interface():
    interface = AnyCallInterface()

    serialised = serialise(interface)
    assert serialised == '"any"'

    deserialised = deserialise(serialised, type=CallInterface)
    assert deserialised == interface


@pytest.mark.parametrize(
    "args, kwargs",
    [
        ((), {}),
        (("a",), {}),
        (("a", "b"), {}),
        ((), {"A": "a"}),
        ((), {"A": "a", "B": "b"}),
        (("a", "b"), {"C": "c", "D": "d"}),
    ],
)
def test_call_arguments(args: tuple[str, ...], kwargs: dict[str, str]):
    call_arguments = CallArguments(args=args, kwargs=kwargs)

    serialised = serialise(call_arguments)
    assert json.loads(serialised) == {
        "args": list(args),
        "kwargs": kwargs,
    }

    deserialised = deserialise(serialised, type=CallArguments)
    assert deserialised == call_arguments

    assert isinstance(deserialised.args, tuple)
    assert isinstance(deserialised.kwargs, frozendict)


def test_location(os_dependent_path: OsDependentPathFn):
    location = Location(
        lineno=123,
        col_offset=50,
        end_lineno=128,
        end_col_offset=1,
        file=Path("/path/to/the/file.py"),
    )

    serialised = serialise(location)
    assert json.loads(serialised) == {
        "lineno": 123,
        "end_lineno": 128,
        "col_offset": 50,
        "end_col_offset": 1,
        "file": os_dependent_path("/path/to/the/file.py"),
    }

    deserialised = deserialise(serialised, type=Location)
    assert deserialised == location
    assert isinstance(deserialised.file, Path)


def test_location_ast_from_token(os_dependent_path: OsDependentPathFn):
    token = ast.parse("x = 123").body[0]
    location = Location.from_ast_token(token, file=Path("/path/to/the/file.py"))

    serialised = serialise(location)
    assert json.loads(serialised) == {
        "lineno": 1,
        "end_lineno": 1,
        "col_offset": 0,
        "end_col_offset": 7,
        "file": os_dependent_path("/path/to/the/file.py"),
    }

    deserialised = deserialise(serialised, type=Location)
    assert deserialised == location


def test_the_empty_symbol_table(make_symbol_table: MakeSymbolTableFn):
    symbol_table = make_symbol_table([], include_root_symbols=False)
    assert symbol_table._symbols == {}

    serialised = serialise(symbol_table)
    assert json.loads(serialised) == {}

    deserialised = deserialise(serialised, type=SymbolTable)
    assert deserialised == symbol_table


def test_symbol_table(
    make_symbol_table: MakeSymbolTableFn,
    symbol_group_one: list[Symbol],
    os_dependent_path: OsDependentPathFn,
):
    symbol_table = make_symbol_table(symbol_group_one, include_root_symbols=False)
    assert symbol_table._symbols != {}

    serialised = serialise(symbol_table)
    assert json.loads(serialised) == {
        "bob": {
            "type": "Name",
            "name": "bob",
            "basename": "bob",
            "location": {
                "lineno": 1,
                "col_offset": 0,
                "end_lineno": None,
                "end_col_offset": None,
                "file": os_dependent_path("test.py"),
            },
            "interface": None,
        },
        "my_func": {
            "type": "Func",
            "name": "my_func",
            "location": {
                "lineno": 1,
                "col_offset": 0,
                "end_lineno": None,
                "end_col_offset": None,
                "file": os_dependent_path("test.py"),
            },
            "interface": {
                "posonlyargs": ["a", "b"],
                "args": [],
                "vararg": None,
                "kwonlyargs": [],
                "kwarg": None,
            },
            "is_async": False,
        },
        "foo.bar.baz": {
            "type": "Name",
            "name": "foo.bar.baz",
            "basename": "foo",
            "location": {
                "lineno": 1,
                "col_offset": 0,
                "end_lineno": None,
                "end_col_offset": None,
                "file": os_dependent_path("test.py"),
            },
            "interface": None,
        },
    }

    deserialised = deserialise(serialised, type=SymbolTable)
    assert deserialised == symbol_table


def test_the_empty_context(
    make_root_context: MakeRootContextFn,
    os_dependent_path: OsDependentPathFn,
):
    context = make_root_context([], include_root_symbols=False)
    assert context.symbol_table._symbols == {}

    serialised = serialise(context)
    assert json.loads(serialised) == {
        "parent": None,
        "symbol_table": {},
        "file": os_dependent_path("test.py"),
    }

    deserialised = deserialise(serialised, type=Context)
    assert deserialised == context


def test_context(
    simple_context_one: Context,
    os_dependent_path: OsDependentPathFn,
):
    assert simple_context_one.symbol_table._symbols != {}

    serialised = serialise(simple_context_one)
    assert json.loads(serialised) == {
        "parent": None,
        "symbol_table": {
            "bob": {
                "type": "Name",
                "name": "bob",
                "basename": "bob",
                "location": {
                    "lineno": 1,
                    "col_offset": 0,
                    "end_lineno": None,
                    "end_col_offset": None,
                    "file": os_dependent_path("test.py"),
                },
                "interface": None,
            },
            "my_func": {
                "type": "Func",
                "name": "my_func",
                "location": {
                    "lineno": 1,
                    "col_offset": 0,
                    "end_lineno": None,
                    "end_col_offset": None,
                    "file": os_dependent_path("test.py"),
                },
                "interface": {
                    "posonlyargs": ["a", "b"],
                    "args": [],
                    "vararg": None,
                    "kwonlyargs": [],
                    "kwarg": None,
                },
                "is_async": False,
            },
            "foo.bar.baz": {
                "type": "Name",
                "name": "foo.bar.baz",
                "basename": "foo",
                "location": {
                    "lineno": 1,
                    "col_offset": 0,
                    "end_lineno": None,
                    "end_col_offset": None,
                    "file": os_dependent_path("test.py"),
                },
                "interface": None,
            },
        },
        "file": os_dependent_path("test.py"),
    }

    deserialised = deserialise(serialised, type=Context)
    assert deserialised == simple_context_one


def test_context_with_parent(
    nested_context: Context,
    os_dependent_path: OsDependentPathFn,
):
    serialised = serialise(nested_context)
    assert json.loads(serialised) == {
        "parent": {
            "parent": None,
            "symbol_table": {
                "bob": {
                    "type": "Name",
                    "name": "bob",
                    "basename": "bob",
                    "location": {
                        "lineno": 1,
                        "col_offset": 0,
                        "end_lineno": None,
                        "end_col_offset": None,
                        "file": os_dependent_path("test.py"),
                    },
                    "interface": None,
                },
                "my_func": {
                    "type": "Func",
                    "name": "my_func",
                    "location": {
                        "lineno": 1,
                        "col_offset": 0,
                        "end_lineno": None,
                        "end_col_offset": None,
                        "file": os_dependent_path("test.py"),
                    },
                    "interface": {
                        "posonlyargs": ["a", "b"],
                        "args": [],
                        "vararg": None,
                        "kwonlyargs": [],
                        "kwarg": None,
                    },
                    "is_async": False,
                },
                "foo.bar.baz": {
                    "type": "Name",
                    "name": "foo.bar.baz",
                    "basename": "foo",
                    "location": {
                        "lineno": 1,
                        "col_offset": 0,
                        "end_lineno": None,
                        "end_col_offset": None,
                        "file": os_dependent_path("test.py"),
                    },
                    "interface": None,
                },
            },
            "file": os_dependent_path("test.py"),
        },
        "symbol_table": {
            "i.am.in.the.child.context": {
                "type": "Name",
                "name": "i.am.in.the.child.context",
                "basename": "i",
                "location": {
                    "lineno": 1,
                    "col_offset": 0,
                    "end_lineno": None,
                    "end_col_offset": None,
                    "file": os_dependent_path("test.py"),
                },
                "interface": None,
            },
            "a_function_here": {
                "type": "Func",
                "name": "a_function_here",
                "location": {
                    "lineno": 1,
                    "col_offset": 0,
                    "end_lineno": None,
                    "end_col_offset": None,
                    "file": os_dependent_path("test.py"),
                },
                "interface": {
                    "posonlyargs": [],
                    "args": ["x", "y"],
                    "vararg": "extras",
                    "kwonlyargs": [],
                    "kwarg": None,
                },
                "is_async": False,
            },
        },
        "file": os_dependent_path("/some/other/file.py"),
    }

    deserialised = deserialise(serialised, type=Context)
    assert deserialised == nested_context


def test_the_empty_file_ir():
    file_ir = FileIr(context=None, file_ir={})

    serialised = serialise(file_ir)
    assert json.loads(serialised) == {
        "context": None,
        "symbols": {},
        "function_irs": {},
    }

    deserialised = deserialise(serialised, type=FileIr)
    assert deserialised == file_ir


def test_file_ir(
    nested_context: Context,
    full_file_ir: dict[UserDefinedCallableSymbol, FunctionIr],
    os_dependent_path: OsDependentPathFn,
):
    file_ir = FileIr(context=nested_context, file_ir=full_file_ir)

    serialised = serialise(file_ir)
    assert json.loads(serialised) == {
        "context": {
            "parent": {
                "parent": None,
                "symbol_table": {
                    "bob": {
                        "type": "Name",
                        "name": "bob",
                        "basename": "bob",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                    "my_func": {
                        "type": "Func",
                        "name": "my_func",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": {
                            "posonlyargs": ["a", "b"],
                            "args": [],
                            "vararg": None,
                            "kwonlyargs": [],
                            "kwarg": None,
                        },
                        "is_async": False,
                    },
                    "foo.bar.baz": {
                        "type": "Name",
                        "name": "foo.bar.baz",
                        "basename": "foo",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                },
                "file": os_dependent_path("test.py"),
            },
            "symbol_table": {
                "i.am.in.the.child.context": {
                    "type": "Name",
                    "name": "i.am.in.the.child.context",
                    "basename": "i",
                    "location": {
                        "lineno": 1,
                        "col_offset": 0,
                        "end_lineno": None,
                        "end_col_offset": None,
                        "file": os_dependent_path("test.py"),
                    },
                    "interface": None,
                },
                "a_function_here": {
                    "type": "Func",
                    "name": "a_function_here",
                    "location": {
                        "lineno": 1,
                        "col_offset": 0,
                        "end_lineno": None,
                        "end_col_offset": None,
                        "file": os_dependent_path("test.py"),
                    },
                    "interface": {
                        "posonlyargs": [],
                        "args": ["x", "y"],
                        "vararg": "extras",
                        "kwonlyargs": [],
                        "kwarg": None,
                    },
                    "is_async": False,
                },
            },
            "file": os_dependent_path("/some/other/file.py"),
        },
        "symbols": {
            "my_func": {
                "type": "Func",
                "name": "my_func",
                "location": {
                    "lineno": 1,
                    "col_offset": 0,
                    "end_lineno": None,
                    "end_col_offset": None,
                    "file": os_dependent_path("test.py"),
                },
                "interface": {
                    "posonlyargs": ["a", "b"],
                    "args": [],
                    "vararg": None,
                    "kwonlyargs": [],
                    "kwarg": None,
                },
                "is_async": False,
            },
            "a_function_here": {
                "type": "Func",
                "name": "a_function_here",
                "location": {
                    "lineno": 1,
                    "col_offset": 0,
                    "end_lineno": None,
                    "end_col_offset": None,
                    "file": os_dependent_path("test.py"),
                },
                "interface": {
                    "posonlyargs": [],
                    "args": ["x", "y"],
                    "vararg": "extras",
                    "kwonlyargs": [],
                    "kwarg": None,
                },
                "is_async": False,
            },
            "MyClass": {
                "type": "Class",
                "name": "MyClass",
                "location": {
                    "lineno": 1,
                    "col_offset": 0,
                    "end_lineno": None,
                    "end_col_offset": None,
                    "file": os_dependent_path("test.py"),
                },
                "interface": {
                    "posonlyargs": [],
                    "args": ["self", "data"],
                    "vararg": None,
                    "kwonlyargs": ["kwarg_a", "kwarg_b"],
                    "kwarg": None,
                },
            },
        },
        "function_irs": {
            "MyClass": {
                "gets": [
                    {
                        "type": "Name",
                        "name": "data",
                        "basename": "data",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                    {
                        "type": "Name",
                        "name": "kwarg_a",
                        "basename": "kwarg_a",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                    {
                        "type": "Name",
                        "name": "kwarg_b",
                        "basename": "kwarg_b",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                    {
                        "type": "Name",
                        "name": "self",
                        "basename": "self",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                ],
                "sets": [
                    {
                        "type": "Name",
                        "name": "self._a",
                        "basename": "self",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                    {
                        "type": "Name",
                        "name": "self._b",
                        "basename": "self",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                    {
                        "type": "Name",
                        "name": "self.data",
                        "basename": "self",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                ],
                "dels": [],
                "calls": [],
            },
            "a_function_here": {
                "gets": [
                    {
                        "type": "Name",
                        "name": "foo",
                        "basename": "foo",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                    {
                        "type": "Name",
                        "name": "foo.bar.baz",
                        "basename": "foo",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                ],
                "sets": [],
                "dels": [],
                "calls": [
                    {
                        "type": "Call",
                        "name": "my_func",
                        "args": {"args": ["foo", "foo"], "kwargs": {}},
                        "target": {
                            "type": "Func",
                            "name": "my_func",
                            "location": {
                                "lineno": 1,
                                "col_offset": 0,
                                "end_lineno": None,
                                "end_col_offset": None,
                                "file": os_dependent_path("test.py"),
                            },
                            "interface": {
                                "posonlyargs": ["a", "b"],
                                "args": [],
                                "vararg": None,
                                "kwonlyargs": [],
                                "kwarg": None,
                            },
                            "is_async": False,
                        },
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                    }
                ],
            },
            "my_func": {
                "gets": [
                    {
                        "type": "Name",
                        "name": "a",
                        "basename": "a",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                    {
                        "type": "Name",
                        "name": "b.bar.baz",
                        "basename": "b",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": None,
                    },
                ],
                "sets": [],
                "dels": [],
                "calls": [],
            },
        },
    }

    deserialised = deserialise(serialised, type=FileIr)
    assert deserialised == file_ir


def test_the_empty_output_irs():
    output_irs = OutputIrs(
        import_irs={},
        target_ir={"filename": "file.py", "ir": FileIr(context=None, file_ir={})},
    )

    serialised = serialise_irs(
        target_name="file.py",
        target_ir=FileIr(context=None, file_ir={}),
        import_irs={},
    )
    assert json.loads(serialised) == {
        "import_irs": {},
        "target_ir": {
            "filename": "file.py",
            "ir": {"context": None, "function_irs": {}, "symbols": {}},
        },
    }

    deserialised = deserialise(serialised, type=OutputIrs)
    assert deserialised == output_irs


def test_output_irs(
    full_file_ir: dict[UserDefinedCallableSymbol, FunctionIr],
    nested_context: Context,
    simple_context_one: Context,
    simple_context_two: Context,
    os_dependent_path: OsDependentPathFn,
):
    output_irs = OutputIrs(
        import_irs={
            "some.module": FileIr(
                context=simple_context_one,
                file_ir={
                    (
                        my_func := Func(
                            "my_func",
                            interface=CallInterface(posonlyargs=("a", "b")),
                        )
                    ): FunctionIr.new(
                        gets={
                            Name("a"),
                            Name("a.attr"),
                            Name("b.another_attr"),
                            Name("a.thing"),
                        },
                        sets={
                            Name("local"),
                        },
                    ),
                    Func(
                        "my_func_two",
                        interface=CallInterface(posonlyargs=("arg",)),
                    ): FunctionIr.new(
                        gets={
                            Name("arg"),
                            Name("arg.attr"),
                            Name("arg.another_attr"),
                            Name("arg.thing"),
                        },
                        sets={
                            Name("local"),
                        },
                        calls={
                            Call(
                                name="my_func",
                                args=CallArguments(args=("arg", "arg")),
                                target=my_func,
                            )
                        },
                    ),
                },
            ),
            "another.module.here": FileIr(
                context=simple_context_two,
                file_ir={
                    Func(
                        "a_function_here",
                        interface=CallInterface(args=("x", "y"), vararg="extras"),
                    ): FunctionIr.new(
                        gets={
                            Name("x"),
                            Name("y"),
                            Name("x.a.b.c.d"),
                        },
                    ),
                },
            ),
            "an_empty_module": FileIr(context=None, file_ir={}),
        },
        target_ir={
            "filename": "file.py",
            "ir": FileIr(context=nested_context, file_ir=full_file_ir),
        },
    )

    serialised = serialise_irs(
        target_name=output_irs.target_ir["filename"],
        target_ir=output_irs.target_ir["ir"],
        import_irs=output_irs.import_irs,
    )
    assert json.loads(serialised) == {
        "import_irs": {
            "some.module": {
                "context": {
                    "parent": None,
                    "symbol_table": {
                        "bob": {
                            "type": "Name",
                            "name": "bob",
                            "basename": "bob",
                            "location": {
                                "lineno": 1,
                                "col_offset": 0,
                                "end_lineno": None,
                                "end_col_offset": None,
                                "file": os_dependent_path("test.py"),
                            },
                            "interface": None,
                        },
                        "my_func": {
                            "type": "Func",
                            "name": "my_func",
                            "location": {
                                "lineno": 1,
                                "col_offset": 0,
                                "end_lineno": None,
                                "end_col_offset": None,
                                "file": os_dependent_path("test.py"),
                            },
                            "interface": {
                                "posonlyargs": ["a", "b"],
                                "args": [],
                                "vararg": None,
                                "kwonlyargs": [],
                                "kwarg": None,
                            },
                            "is_async": False,
                        },
                        "foo.bar.baz": {
                            "type": "Name",
                            "name": "foo.bar.baz",
                            "basename": "foo",
                            "location": {
                                "lineno": 1,
                                "col_offset": 0,
                                "end_lineno": None,
                                "end_col_offset": None,
                                "file": os_dependent_path("test.py"),
                            },
                            "interface": None,
                        },
                    },
                    "file": os_dependent_path("test.py"),
                },
                "symbols": {
                    "my_func": {
                        "type": "Func",
                        "name": "my_func",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": {
                            "posonlyargs": ["a", "b"],
                            "args": [],
                            "vararg": None,
                            "kwonlyargs": [],
                            "kwarg": None,
                        },
                        "is_async": False,
                    },
                    "my_func_two": {
                        "type": "Func",
                        "name": "my_func_two",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": {
                            "posonlyargs": ["arg"],
                            "args": [],
                            "vararg": None,
                            "kwonlyargs": [],
                            "kwarg": None,
                        },
                        "is_async": False,
                    },
                },
                "function_irs": {
                    "my_func": {
                        "gets": [
                            {
                                "type": "Name",
                                "name": "a",
                                "basename": "a",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "a.attr",
                                "basename": "a",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "a.thing",
                                "basename": "a",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "b.another_attr",
                                "basename": "b",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                        ],
                        "sets": [
                            {
                                "type": "Name",
                                "name": "local",
                                "basename": "local",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            }
                        ],
                        "dels": [],
                        "calls": [],
                    },
                    "my_func_two": {
                        "gets": [
                            {
                                "type": "Name",
                                "name": "arg",
                                "basename": "arg",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "arg.another_attr",
                                "basename": "arg",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "arg.attr",
                                "basename": "arg",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "arg.thing",
                                "basename": "arg",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                        ],
                        "sets": [
                            {
                                "type": "Name",
                                "name": "local",
                                "basename": "local",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            }
                        ],
                        "dels": [],
                        "calls": [
                            {
                                "type": "Call",
                                "name": "my_func",
                                "args": {"args": ["arg", "arg"], "kwargs": {}},
                                "target": {
                                    "type": "Func",
                                    "name": "my_func",
                                    "location": {
                                        "lineno": 1,
                                        "col_offset": 0,
                                        "end_lineno": None,
                                        "end_col_offset": None,
                                        "file": os_dependent_path("test.py"),
                                    },
                                    "interface": {
                                        "posonlyargs": ["a", "b"],
                                        "args": [],
                                        "vararg": None,
                                        "kwonlyargs": [],
                                        "kwarg": None,
                                    },
                                    "is_async": False,
                                },
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                            }
                        ],
                    },
                },
            },
            "another.module.here": {
                "context": {
                    "parent": None,
                    "symbol_table": {
                        "i.am.in.the.child.context": {
                            "type": "Name",
                            "name": "i.am.in.the.child.context",
                            "basename": "i",
                            "location": {
                                "lineno": 1,
                                "col_offset": 0,
                                "end_lineno": None,
                                "end_col_offset": None,
                                "file": os_dependent_path("test.py"),
                            },
                            "interface": None,
                        },
                        "a_function_here": {
                            "type": "Func",
                            "name": "a_function_here",
                            "location": {
                                "lineno": 1,
                                "col_offset": 0,
                                "end_lineno": None,
                                "end_col_offset": None,
                                "file": os_dependent_path("test.py"),
                            },
                            "interface": {
                                "posonlyargs": [],
                                "args": ["x", "y"],
                                "vararg": "extras",
                                "kwonlyargs": [],
                                "kwarg": None,
                            },
                            "is_async": False,
                        },
                    },
                    "file": os_dependent_path("test.py"),
                },
                "symbols": {
                    "a_function_here": {
                        "type": "Func",
                        "name": "a_function_here",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": {
                            "posonlyargs": [],
                            "args": ["x", "y"],
                            "vararg": "extras",
                            "kwonlyargs": [],
                            "kwarg": None,
                        },
                        "is_async": False,
                    }
                },
                "function_irs": {
                    "a_function_here": {
                        "gets": [
                            {
                                "type": "Name",
                                "name": "x",
                                "basename": "x",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "x.a.b.c.d",
                                "basename": "x",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "y",
                                "basename": "y",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                        ],
                        "sets": [],
                        "dels": [],
                        "calls": [],
                    }
                },
            },
            "an_empty_module": {"context": None, "symbols": {}, "function_irs": {}},
        },
        "target_ir": {
            "filename": "file.py",
            "ir": {
                "context": {
                    "parent": {
                        "parent": None,
                        "symbol_table": {
                            "bob": {
                                "type": "Name",
                                "name": "bob",
                                "basename": "bob",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            "my_func": {
                                "type": "Func",
                                "name": "my_func",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": {
                                    "posonlyargs": ["a", "b"],
                                    "args": [],
                                    "vararg": None,
                                    "kwonlyargs": [],
                                    "kwarg": None,
                                },
                                "is_async": False,
                            },
                            "foo.bar.baz": {
                                "type": "Name",
                                "name": "foo.bar.baz",
                                "basename": "foo",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                        },
                        "file": os_dependent_path("test.py"),
                    },
                    "symbol_table": {
                        "i.am.in.the.child.context": {
                            "type": "Name",
                            "name": "i.am.in.the.child.context",
                            "basename": "i",
                            "location": {
                                "lineno": 1,
                                "col_offset": 0,
                                "end_lineno": None,
                                "end_col_offset": None,
                                "file": os_dependent_path("test.py"),
                            },
                            "interface": None,
                        },
                        "a_function_here": {
                            "type": "Func",
                            "name": "a_function_here",
                            "location": {
                                "lineno": 1,
                                "col_offset": 0,
                                "end_lineno": None,
                                "end_col_offset": None,
                                "file": os_dependent_path("test.py"),
                            },
                            "interface": {
                                "posonlyargs": [],
                                "args": ["x", "y"],
                                "vararg": "extras",
                                "kwonlyargs": [],
                                "kwarg": None,
                            },
                            "is_async": False,
                        },
                    },
                    "file": os_dependent_path("/some/other/file.py"),
                },
                "symbols": {
                    "MyClass": {
                        "type": "Class",
                        "name": "MyClass",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": {
                            "posonlyargs": [],
                            "args": ["self", "data"],
                            "vararg": None,
                            "kwonlyargs": ["kwarg_a", "kwarg_b"],
                            "kwarg": None,
                        },
                    },
                    "a_function_here": {
                        "type": "Func",
                        "name": "a_function_here",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": {
                            "posonlyargs": [],
                            "args": ["x", "y"],
                            "vararg": "extras",
                            "kwonlyargs": [],
                            "kwarg": None,
                        },
                        "is_async": False,
                    },
                    "my_func": {
                        "type": "Func",
                        "name": "my_func",
                        "location": {
                            "lineno": 1,
                            "col_offset": 0,
                            "end_lineno": None,
                            "end_col_offset": None,
                            "file": os_dependent_path("test.py"),
                        },
                        "interface": {
                            "posonlyargs": ["a", "b"],
                            "args": [],
                            "vararg": None,
                            "kwonlyargs": [],
                            "kwarg": None,
                        },
                        "is_async": False,
                    },
                },
                "function_irs": {
                    "MyClass": {
                        "gets": [
                            {
                                "type": "Name",
                                "name": "data",
                                "basename": "data",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "kwarg_a",
                                "basename": "kwarg_a",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "kwarg_b",
                                "basename": "kwarg_b",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "self",
                                "basename": "self",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                        ],
                        "sets": [
                            {
                                "type": "Name",
                                "name": "self._a",
                                "basename": "self",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "self._b",
                                "basename": "self",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "self.data",
                                "basename": "self",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                        ],
                        "dels": [],
                        "calls": [],
                    },
                    "a_function_here": {
                        "gets": [
                            {
                                "type": "Name",
                                "name": "foo",
                                "basename": "foo",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "foo.bar.baz",
                                "basename": "foo",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                        ],
                        "sets": [],
                        "dels": [],
                        "calls": [
                            {
                                "type": "Call",
                                "name": "my_func",
                                "args": {"args": ["foo", "foo"], "kwargs": {}},
                                "target": {
                                    "type": "Func",
                                    "name": "my_func",
                                    "location": {
                                        "lineno": 1,
                                        "col_offset": 0,
                                        "end_lineno": None,
                                        "end_col_offset": None,
                                        "file": os_dependent_path("test.py"),
                                    },
                                    "interface": {
                                        "posonlyargs": ["a", "b"],
                                        "args": [],
                                        "vararg": None,
                                        "kwonlyargs": [],
                                        "kwarg": None,
                                    },
                                    "is_async": False,
                                },
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                            }
                        ],
                    },
                    "my_func": {
                        "gets": [
                            {
                                "type": "Name",
                                "name": "a",
                                "basename": "a",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                            {
                                "type": "Name",
                                "name": "b.bar.baz",
                                "basename": "b",
                                "location": {
                                    "lineno": 1,
                                    "col_offset": 0,
                                    "end_lineno": None,
                                    "end_col_offset": None,
                                    "file": os_dependent_path("test.py"),
                                },
                                "interface": None,
                            },
                        ],
                        "sets": [],
                        "dels": [],
                        "calls": [],
                    },
                },
            },
        },
    }

    deserialised = deserialise(serialised, type=OutputIrs)
    assert deserialised == output_irs


def test_the_empty_file_results():
    results = FileResults({})

    serialised = serialise(results)
    assert json.loads(serialised) == {}

    deserialised = deserialise(serialised, type=FileResults)
    assert deserialised == results


def test_file_results():
    results = FileResults(
        {
            "foo": FunctionResults.the_empty_results(),
            "bar": FunctionResults.new(
                gets={"arg", "arg.attr"},
                calls={"foo"},
            ),
            "baz": FunctionResults.new(
                gets={
                    "x",
                    "x.attr",
                    "y",
                    "y.attr",
                    "x.some_flag",
                    "y.set_me",
                },
                sets={
                    "y.set_me",
                },
                dels={
                    "del_me",
                },
                calls={
                    "foo",
                    "bar",
                },
            ),
        }
    )

    serialised = serialise(results)
    assert json.loads(serialised) == {
        "bar": {"gets": ["arg", "arg.attr"], "sets": [], "dels": [], "calls": ["foo"]},
        "baz": {
            "gets": ["x", "x.attr", "x.some_flag", "y", "y.attr", "y.set_me"],
            "sets": ["y.set_me"],
            "dels": ["del_me"],
            "calls": ["bar", "foo"],
        },
        "foo": {"gets": [], "sets": [], "dels": [], "calls": []},
    }

    deserialised = deserialise(serialised, type=FileResults)
    assert deserialised == results



from __future__ import annotations

from typing import TYPE_CHECKING

from rattr import error
from rattr.config.state import enter_file

if TYPE_CHECKING:
    import ast


class TestErrorUtils:
    def test_get_file_and_line_info(self, parse):
        ast_module: ast.Module = parse(
            """
            # not on line 1!
            # not line 2,
            x = 4
            # but third time's the charm
            """
        )
        culprit: ast.expr = ast_module.body[0]

        line_info = "\033[1m:{}:{}\033[0m".format(culprit.lineno, culprit.col_offset)

        # No file, no culprit
        # No file, w/ culprit
        with enter_file(None):
            file_info = "\033[1m{}\033[0m".format("target.py")
            assert error.get_file_and_line_info(None) == (file_info, "")
            assert error.get_file_and_line_info(culprit) == (file_info, line_info)

        # File, no culprit
        # File, w/ culprit
        with enter_file(file := "some_file_name.py"):
            file_info = "\033[1m{}\033[0m".format(file)
            assert error.get_file_and_line_info(None) == (file_info, "")
            assert error.get_file_and_line_info(culprit) == (file_info, line_info)
